from torch.utils.data.sampler import Sampler
import random
import numpy as np


class MyResampleSampler(Sampler):
    def __init__(self, data, id_index_dict_pos, id_index_dict_mid, id_index_dict_neg, resample_num_pos,
                 resample_num_neg, resample_num_mid,
                 batch_size_all=None, shuffle_type=None, dataset_id_map=None, only_mixup_bad_particles=False,
                 error_balance=None, mean_error_dis_dict=None, data_error_dis_dict=None,
                 balance_per_interval=False,
                 index_list=[0.3]
                 ):
        self.data = data
        self.id_index_dict_pos = id_index_dict_pos
        self.id_index_dict_mid = id_index_dict_mid
        self.id_index_dict_neg = id_index_dict_neg
        self.resample_num_pos = resample_num_pos
        self.resample_num_mid = resample_num_mid
        self.resample_num_neg = resample_num_neg
        self.shuffle_type = shuffle_type
        self.batch_size_all = batch_size_all
        self.dataset_id_map = dataset_id_map
        self.my_seed = 0
        self.only_mixup_bad_particles = only_mixup_bad_particles
        self.error_balance = error_balance
        self.mean_error_dis_dict = mean_error_dis_dict
        self.data_error_dis_dict = data_error_dis_dict
        self.balance_per_interval = balance_per_interval
        self.index_list = index_list
        self.indices = resample_from_id_index_dict_finetune(self.id_index_dict_pos,
                                                            self.id_index_dict_mid,
                                                            self.id_index_dict_neg,
                                                            self.resample_num_pos,
                                                            self.resample_num_mid,
                                                            self.resample_num_neg, batch_size_all=self.batch_size_all,
                                                            shuffle_type=self.shuffle_type, my_seed=self.my_seed,
                                                            dataset_id_map=dataset_id_map,
                                                            only_mixup_bad_particles=only_mixup_bad_particles,
                                                            balance_per_interval=balance_per_interval,
                                                            index_list=index_list
                                                            # , error_balance=error_balance,
                                                            # mean_error_dis_dict=mean_error_dis_dict
                                                            # , data_error_dis_dict=data_error_dis_dict
                                                            )

    def __iter__(self):
        self.indices = resample_from_id_index_dict_finetune(self.id_index_dict_pos,
                                                            self.id_index_dict_mid,
                                                            self.id_index_dict_neg,
                                                            self.resample_num_pos,
                                                            self.resample_num_mid,
                                                            self.resample_num_neg, batch_size_all=self.batch_size_all,
                                                            shuffle_type=self.shuffle_type, my_seed=self.my_seed,
                                                            dataset_id_map=self.dataset_id_map,
                                                            only_mixup_bad_particles=self.only_mixup_bad_particles,
                                                            balance_per_interval=self.balance_per_interval,
                                                            index_list=self.index_list,
                                                            # , error_balance=self.error_balance,
                                                            # mean_error_dis_dict=self.mean_error_dis_dict
                                                            # , data_error_dis_dict=self.data_error_dis_dict
                                                            )
        self.my_seed += 1
        return iter(self.indices)

    def __len__(self):
        # return len(self.data)
        return len(self.indices)


class MyResampleSampler_pretrain(Sampler):
    def __init__(self, id_index_dict, batch_size_all, max_number_per_sample=None, shuffle_type=None,
                 shuffle_mix_up_ratio=0.2, dataset_id_map=None, bad_particles_ratio=0.1, combine_same_class=False,
                 only_mixup_bad_particles=False, id_scores_dict=None, scores_bar=0.8):
        self.id_index_dict = id_index_dict
        self.batch_size_all = batch_size_all
        self.max_number_per_sample = max_number_per_sample
        self.shuffle_type = shuffle_type
        self.shuffle_mix_up_ratio = shuffle_mix_up_ratio
        self.my_seed = 0
        self.dataset_id_map = dataset_id_map
        self.bad_particles_ratio = bad_particles_ratio
        self.combine_same_class = combine_same_class
        self.only_mixup_bad_particles = only_mixup_bad_particles
        self.id_scores_dict = id_scores_dict
        self.scores_bar = scores_bar
        self.indices = resample_from_id_index_dict(id_index_dict, max_number_per_sample, batch_size_all, shuffle_type,
                                                   shuffle_mix_up_ratio, self.my_seed, dataset_id_map,
                                                   bad_particles_ratio=bad_particles_ratio,
                                                   combine_same_class=combine_same_class,
                                                   only_mixup_bad_particles=only_mixup_bad_particles,
                                                   id_scores_dict=id_scores_dict,
                                                   scores_bar=scores_bar
                                                   )

    # @profile(precision=4)
    def __iter__(self):
        self.indices = resample_from_id_index_dict(self.id_index_dict, self.max_number_per_sample, self.batch_size_all,
                                                   self.shuffle_type, self.shuffle_mix_up_ratio, self.my_seed,
                                                   self.dataset_id_map, bad_particles_ratio=self.bad_particles_ratio,
                                                   combine_same_class=self.combine_same_class,
                                                   only_mixup_bad_particles=self.only_mixup_bad_particles,
                                                   id_scores_dict=self.id_scores_dict,
                                                   scores_bar=self.scores_bar
                                                   )
        # print(self.indices[0:80])
        self.my_seed += 1
        # print(sorted(self.indices))
        return iter(self.indices)

    def __len__(self):
        # return len(self.data)
        return len(self.indices)


class MyResampleSampler_pretrain_old(Sampler):
    def __init__(self, id_index_dict, max_number_per_sample, bad_particles_id=[]):
        self.id_index_dict = id_index_dict
        self.max_number_per_sample = max_number_per_sample
        self.indices = resample_from_id_index_dict_old(id_index_dict, max_number_per_sample, bad_particles_id)
        self.bad_particles_id = bad_particles_id

    def __iter__(self):
        self.indices = resample_from_id_index_dict_old(self.id_index_dict, self.max_number_per_sample,
                                                       self.bad_particles_id)

        return iter(self.indices)

    def __len__(self):
        # return len(self.data)
        return len(self.indices)


# # Optimize the resample_from_id_index_dict function
def resample_from_id_index_dict(id_index_dict, max_number_per_sample=None, batch_size_all=None, shuffle_type=None,
                                shuffle_mix_up_ratio=0.2, my_seed=0, dataset_id_map=None, bad_particles_ratio=0.1,
                                combine_same_class=False, only_mixup_bad_particles=False, error_balance=None,
                                mean_error_dis_dict=None, data_error_dis_dict=None,
                                id_scores_dict=None, scores_bar=0.0,
                                balance_per_interval=False,
                                interval_list=[0.3]
                                ):
    random.seed(my_seed)
    resampled_index_list = []
    final_resampled_index_list = []
    ids_list = list(id_index_dict.keys())
    mix_up_list = []
    ids_list_all = list(id_index_dict.keys())

    if dataset_id_map is not None:
        id_map = dataset_id_map['id_map']
        bad_id_list = dataset_id_map['bad_id_list']
        bad_id_list_all = dataset_id_map['bad_id_list_all']
        if id_map is not None:
            ids_list = list(id_map.keys())
        else:
            ids_list = None


    else:
        id_map = None
        bad_id_list = []
        bad_id_list_all = []

    # if shuffle_type == 'class':
    #     random.shuffle(ids_list)
    if shuffle_type == 'class':
        if ids_list is not None:
            random.shuffle(ids_list)
        random.shuffle(ids_list_all)

    if id_map is not None and combine_same_class:
        for my_id in ids_list:
            if id_map[my_id] is not None:
                max_number_per_sample_neg = int(max_number_per_sample * bad_particles_ratio)
                max_number_per_sample_pos = int(max_number_per_sample - max_number_per_sample_neg)

                selected_index_list1, mix_up_list_added1 = get_index_per_class(id_index_dict[my_id],
                                                                               max_number_per_sample_pos, shuffle_type,
                                                                               shuffle_mix_up_ratio,
                                                                               is_bad_class=my_id in bad_id_list,
                                                                               error_distribution=data_error_dis_dict[
                                                                                   my_id] if data_error_dis_dict is not None else None,
                                                                               interval_list=interval_list)

                selected_index_list2, mix_up_list_added2 = get_index_per_class(id_index_dict[id_map[my_id]],
                                                                               max_number_per_sample_neg, shuffle_type,
                                                                               shuffle_mix_up_ratio,
                                                                               error_distribution=data_error_dis_dict[
                                                                                   id_map[
                                                                                       my_id]] if data_error_dis_dict is not None else None,
                                                                               interval_list=interval_list)
                new_selected_index_list = selected_index_list1 + selected_index_list2
                # random.shuffle(new_selected_index_list)
                # resampled_index_list.append(new_selected_index_list)
                mix_up_list_added = mix_up_list_added1 + mix_up_list_added2
                # mix_up_list.extend(mix_up_list_added)
            else:
                if my_id in bad_id_list and not only_mixup_bad_particles:
                    max_number_per_sample_i = max_number_per_sample * bad_particles_ratio
                else:
                    max_number_per_sample_i = max_number_per_sample

                new_selected_index_list, mix_up_list_added = get_index_per_class(id_index_dict[my_id],
                                                                                 max_number_per_sample_i, shuffle_type,
                                                                                 shuffle_mix_up_ratio,
                                                                                 is_bad_class=my_id in bad_id_list
                                                                                 ,
                                                                                 error_distribution=data_error_dis_dict[
                                                                                     my_id] if data_error_dis_dict is not None else None,
                                                                                 interval_list=interval_list)
            if only_mixup_bad_particles:
                if my_id in bad_id_list:
                    new_selected_index_list = []
                else:
                    mix_up_list_added = []

            if len(new_selected_index_list) > 0:
                resampled_index_list.append(new_selected_index_list)
            mix_up_list.extend(mix_up_list_added)
    else:
        for my_id in ids_list_all:

            if my_id in bad_id_list_all and not only_mixup_bad_particles:
                max_number_per_sample_i = int(max_number_per_sample * bad_particles_ratio)
            else:
                max_number_per_sample_i = max_number_per_sample

            if error_balance is not None:
                if error_balance['data_mean_error']:
                    max_number_per_sample_i = int(max_number_per_sample_i * mean_error_dis_dict[my_id])

            new_selected_index_list, mix_up_list_added = get_index_per_class(id_index_dict[my_id],
                                                                             max_number_per_sample_i, shuffle_type,
                                                                             shuffle_mix_up_ratio,
                                                                             is_bad_class=my_id in bad_id_list_all,
                                                                             error_distribution=data_error_dis_dict[
                                                                                 my_id] if data_error_dis_dict is not None else None,
                                                                             # dataset_scores=np.array(id_scores_dict)[id_index_dict[my_id]] if id_scores_dict is not None else None,
                                                                             dataset_scores=id_scores_dict[
                                                                                 my_id] if id_scores_dict is not None else None,
                                                                             # bad_particles_ratio=bad_particles_ratio
                                                                             scores_bar=scores_bar,
                                                                             balance_per_interval=balance_per_interval,
                                                                             interval_list=interval_list
                                                                             )
            if only_mixup_bad_particles:
                if my_id in bad_id_list_all:
                    new_selected_index_list = []
                else:
                    mix_up_list_added = []

            if len(new_selected_index_list) > 0:
                resampled_index_list.append(new_selected_index_list)
            mix_up_list.extend(mix_up_list_added)

    if shuffle_type == 'batch':
        random.shuffle(mix_up_list)
        step = len(mix_up_list) // len(resampled_index_list)
        for i in range(len(resampled_index_list)):
            # step=batch_size_all-len(resampled_index_list[i])
            resampled_index_list[i].extend(mix_up_list[:step])
            random.shuffle(resampled_index_list[i])
            new_resampled_index_list_i = []
            for ii in range(len(resampled_index_list[i]) // batch_size_all + 1):
                if len(resampled_index_list[i]) >= batch_size_all:
                    new_resampled_index_list_i.append(resampled_index_list[i][:batch_size_all])
                    resampled_index_list[i] = resampled_index_list[i][batch_size_all:]
            final_resampled_index_list.extend(new_resampled_index_list_i)
            mix_up_list = mix_up_list[step:]

    if shuffle_type == 'class':
        random.shuffle(mix_up_list)
        step = len(mix_up_list) // len(resampled_index_list)
        for i in range(len(resampled_index_list)):
            # step=batch_size_all-len(resampled_index_list[i])
            resampled_index_list[i].extend(mix_up_list[:step])
            random.shuffle(resampled_index_list[i])
            mix_up_list = mix_up_list[step:]

    if shuffle_type == 'batch':
        random.shuffle(final_resampled_index_list)
        final_resampled_index_list = [item for sublist in final_resampled_index_list for item in sublist]
    else:
        final_resampled_index_list = [item for sublist in resampled_index_list for item in sublist]
        if shuffle_type == 'all':
            random.shuffle(final_resampled_index_list)
    return final_resampled_index_list


def resample_from_id_index_dict_old(id_index_dict, max_number_per_sample, bad_particles_id=[]):
    resampled_index_list = []
    for id, index_list in id_index_dict.items():
        if id in bad_particles_id:
            new_max_number_per_sample = max_number_per_sample // 10
            # new_max_number_per_sample=0
        else:
            new_max_number_per_sample = max_number_per_sample
        if len(index_list) > new_max_number_per_sample:
            resampled_index_list.extend(random.sample(index_list, new_max_number_per_sample))
        else:
            resampled_index_list.extend(index_list)
    return resampled_index_list


def get_index_per_class(index_list, max_number_per_sample=None, shuffle_type=None, shuffle_mix_up_ratio=0.2,
                        is_bad_class=False, error_distribution=None, dataset_scores=None, scores_bar=0.8,
                        balance_per_interval=False, interval_list=[0.3]):
    # index_list = id_index_dict[my_id]
    if balance_per_interval:
        index_list = balance_from_scores_interval(20, index_list['score'], index_list['id'], num_min_per_interval=128,
                                                  interval_list=interval_list)
    else:
        if isinstance(index_list, dict):
            index_list = index_list['id']

    if dataset_scores is not None and scores_bar > 0 and max(dataset_scores) != min(dataset_scores):
        index_list_good = np.array(index_list)[np.array(dataset_scores) >= scores_bar].tolist()
        index_list_bad = np.array(index_list)[np.array(dataset_scores) < scores_bar].tolist()
        index_list = index_list_good

    else:
        index_list_bad = None
    len_index_list = len(index_list)
    if max_number_per_sample is not None and len_index_list > max_number_per_sample:
        if error_distribution is not None:
            index_array = np.array(index_list)
            selected_indices = np.random.choice(index_array, size=max_number_per_sample, replace=False,
                                                p=error_distribution)
            selected_index_list = selected_indices.tolist()
        else:
            selected_index_list = random.sample(index_list, int(max_number_per_sample))

        if shuffle_type == 'batch' or shuffle_type == 'class':
            if is_bad_class:
                mix_up_list_added = selected_index_list[:int(len(selected_index_list) * shuffle_mix_up_ratio)]
                new_selected_index_list = []
            elif index_list_bad is not None:
                if len(index_list_bad) < int(len(selected_index_list) * shuffle_mix_up_ratio):
                    mix_up_list_added = index_list_bad
                else:
                    mix_up_list_added = random.sample(index_list_bad,
                                                      int(len(selected_index_list) * shuffle_mix_up_ratio))
                new_selected_index_list = selected_index_list[
                                          :max_number_per_sample - int(len(selected_index_list) * shuffle_mix_up_ratio)]
            else:
                new_selected_index_list = selected_index_list[
                                          :max_number_per_sample - int(len(selected_index_list) * shuffle_mix_up_ratio)]
                mix_up_list_added = (
                    selected_index_list[max_number_per_sample - int(len(selected_index_list) * shuffle_mix_up_ratio):])
        else:
            if is_bad_class and shuffle_mix_up_ratio > 0:
                mix_up_list_added = selected_index_list
                new_selected_index_list = []
            elif index_list_bad is not None and shuffle_mix_up_ratio > 0:
                if len(index_list_bad) < int(len(index_list_bad) * shuffle_mix_up_ratio):
                    mix_up_list_added = index_list_bad
                else:

                    mix_up_list_added = random.sample(index_list_bad, int(len(index_list_bad) * shuffle_mix_up_ratio))
                new_selected_index_list = selected_index_list[
                                          :max_number_per_sample - int(len(index_list_bad) * shuffle_mix_up_ratio)]
            else:
                new_selected_index_list = selected_index_list
                mix_up_list_added = []
    else:
        if is_bad_class:
            mix_up_list_added = index_list[:int(len(index_list) * shuffle_mix_up_ratio)]
            new_selected_index_list = []
        elif shuffle_type == 'batch' or shuffle_type == 'class':
            random.shuffle(index_list)
            new_selected_index_list = index_list[:len_index_list - int(len(index_list) * shuffle_mix_up_ratio)]
            mix_up_list_added = (index_list[len_index_list - int(len(index_list) * shuffle_mix_up_ratio):])
        else:
            if shuffle_type != 'valset':
                random.shuffle(index_list)
            new_selected_index_list = index_list
            mix_up_list_added = []
    return new_selected_index_list, mix_up_list_added


def resample_from_id_index_dict_finetune(id_index_dict_pos, id_index_dict_mid, id_index_dict_neg, resample_num_p,
                                         resample_num_mid, resample_num_n,
                                         batch_size_all=None, shuffle_type=None, shuffle_mix_up_ratio=0.2, my_seed=0,
                                         dataset_id_map=None, only_mixup_bad_particles=False,
                                         balance_per_interval=False,
                                         index_list=[0.3],
                                         # error_balance=None,
                                         # mean_error_dis_dict=None,
                                         # data_error_dis_dict=None
                                         ):
    random.seed(my_seed)
    if shuffle_type == 'batch':
        id_index_dict_all = {**id_index_dict_pos, **id_index_dict_neg}
        resampled_index_list = resample_from_id_index_dict(id_index_dict_all, resample_num_p + resample_num_n,
                                                           batch_size_all, shuffle_type,
                                                           shuffle_mix_up_ratio, my_seed, dataset_id_map=dataset_id_map,
                                                           only_mixup_bad_particles=only_mixup_bad_particles,
                                                           balance_per_interval=balance_per_interval,
                                                           interval_list=index_list
                                                           # error_balance=error_balance,
                                                           # mean_error_dis_dict=mean_error_dis_dict,
                                                           # data_error_dis_dict=data_error_dis_dict
                                                           # positive_ratio=resample_num_p /(resample_num_p + resample_num_n)
                                                           )

    else:
        resampled_index_list = []
        resampled_index_list.extend(
            resample_from_id_index_dict(id_index_dict_pos, resample_num_p, batch_size_all, shuffle_type,
                                        shuffle_mix_up_ratio, my_seed,
                                        only_mixup_bad_particles=only_mixup_bad_particles,
                                        balance_per_interval=balance_per_interval,
                                        interval_list=index_list
                                        # error_balance=error_balance,
                                        # mean_error_dis_dict=mean_error_dis_dict,
                                        # data_error_dis_dict=data_error_dis_dict['good']
                                        ))
        resampled_index_list.extend(
            resample_from_id_index_dict(id_index_dict_neg, resample_num_n, batch_size_all, shuffle_type,
                                        shuffle_mix_up_ratio, my_seed,
                                        only_mixup_bad_particles=only_mixup_bad_particles,
                                        balance_per_interval=balance_per_interval,
                                        interval_list=index_list
                                        # error_balance=error_balance,
                                        # mean_error_dis_dict=mean_error_dis_dict,
                                        # data_error_dis_dict=data_error_dis_dict['bad']
                                        ))
        if len(id_index_dict_mid) > 0:
            resampled_index_list.extend(
                resample_from_id_index_dict(id_index_dict_mid, resample_num_mid, batch_size_all, shuffle_type,
                                            shuffle_mix_up_ratio, my_seed,
                                            only_mixup_bad_particles=only_mixup_bad_particles,
                                            balance_per_interval=balance_per_interval,
                                            interval_list=index_list
                                            # error_balance=error_balance, mean_error_dis_dict=mean_error_dis_dict,
                                            # data_error_dis_dict=data_error_dis_dict['mid']
                                            ))
        if shuffle_type == 'all':
            random.shuffle(resampled_index_list)
    return resampled_index_list


def balance_from_scores_interval(interval_num, scores, ids, num_min_per_interval=0, interval_list=[0.3]):
    """
    Balance the dataset based on scores intervals.
    :param interval_num: Number of intervals to divide the scores into.
    :param scores: List of scores corresponding to each id.
    :param ids: List of ids corresponding to each score.
    :param num_min_per_interval: Minimum number of samples per interval.
    :return: Balanced ids and scores.
    """

    if len(scores) == 0:
        return [], []

    min_score = min(scores)
    max_score = max(scores)
    if min_score == max_score:
        return ids, scores  # No variation in scores, return original lists
    if len(interval_list)>1:
        interval_num = len(interval_list)
    interval_size = (max_score - min_score) / interval_num
    min_interval_len = len(scores)
    scores_np = np.array(scores)
    ids_np = np.array(ids)

    intervals = [[] for _ in range(interval_num)]
    for i in range(interval_num):
        lower_bound = min_score + i * interval_size
        upper_bound = min_score + (i + 1) * interval_size
        mask = (scores_np >= lower_bound) & (scores_np < upper_bound)
        interval_ids = ids_np[mask].tolist()
        intervals[i] = interval_ids
        if len(interval_ids) < min_interval_len:
            min_interval_len = len(interval_ids)

    if min_interval_len > num_min_per_interval:
        num_min_per_interval = min_interval_len
    balanced_ids = []
    # balanced_scores = []
    if interval_list is not None and len(interval_list) > 0:
        if len(interval_list) == 1:
            new_intervals = [1 - interval_list[0] + interval_list[0] / interval_num * i for i in range(interval_num)]
        else:
            new_intervals = interval_list
    else:
        new_intervals = [1 for _ in range(interval_num)]
    for i, interval in enumerate(intervals):
        if len(interval) < int(num_min_per_interval * new_intervals[i]):
            selected_ids = interval
        else:
            selected_ids = random.sample(interval, int(num_min_per_interval * new_intervals[i]))
        balanced_ids.extend(selected_ids)

    balanced_ids = sorted(balanced_ids)  # Sort the final list of ids

    return balanced_ids



from cryosparc.dataset import Dataset
import numpy as np


def generate_cs_file(blob_path, blob_id, image_size, psize, df1, df2, accel_kv=300, cs_mm=2.7, amp_contrast=0.07,
                     phase_shift_rad=0.0, bfactor=0.0):
    data_len = len(blob_path)
    cs_data = Dataset(data_len)
    cs_data.add_fields(
        ['ctf/type', 'ctf/exp_group_id', 'ctf/accel_kv', 'ctf/cs_mm', 'ctf/amp_contrast',
         'ctf/df1_A','ctf/df2_A','ctf/df_angle_rad', 'ctf/phase_shift_rad', 'ctf/scale',
         'ctf/scale_const', 'ctf/shift_A', 'ctf/tilt_A','ctf/trefoil_A', 'ctf/tetra_A',
         'ctf/anisomag', 'ctf/bfactor'],
        ['str', 'int', 'float', 'float', 'float',
         'float', 'float', 'float', 'float', 'float',
         'float', ('float', (2,)),('float', (2,)), ('float', (2,)),('float', (4,)),
         ('float', (4,)), 'float'])
    cs_data.add_fields(
        ['blob/path', 'blob/psize_A', 'blob/idx', 'blob/shape', 'blob/sign','blob/import_sig'],
        ['str', 'float', 'int', ('int', (2,)), 'float',int])

    cs_data['blob/path'] = blob_path
    # cs_data['blob/psize_A'] = [psize] * data_len
    cs_data['ctf/df1_A'] = df1
    cs_data['ctf/df2_A'] = df2
    cs_data['blob/idx'] = blob_id
    cs_data['blob/sign'] = [1.0] * data_len
    cs_data['ctf/scale'] = [1.0] * data_len
    if not isinstance(accel_kv, list) and type(accel_kv) is not np.ndarray:
        cs_data['ctf/accel_kv'] = [accel_kv] * data_len
    else:
        cs_data['ctf/accel_kv'] = accel_kv
    if not isinstance(image_size, list) and type(image_size) is not np.ndarray:
        cs_data['blob/shape'] = [image_size, image_size] * data_len
    else:
        cs_data['blob/shape'] = image_size
    if not isinstance(psize, list) and type(psize) is not np.ndarray:
        cs_data['blob/psize_A'] = [psize] * data_len
    else:
        cs_data['blob/psize_A'] = psize
    if not isinstance(cs_mm, list) and type(cs_mm) is not np.ndarray:
        cs_data['ctf/cs_mm'] = [cs_mm] * data_len
    else:
        cs_data['ctf/cs_mm'] = cs_mm
    if not isinstance(amp_contrast, list) and type(amp_contrast) is not np.ndarray:
        cs_data['ctf/amp_contrast'] = [amp_contrast] * data_len
    else:
        cs_data['ctf/amp_contrast'] = amp_contrast
    if not isinstance(phase_shift_rad, list) and type(phase_shift_rad) is not np.ndarray:
        cs_data['ctf/phase_shift_rad'] = [phase_shift_rad] * data_len
    else:
        cs_data['ctf/phase_shift_rad'] = phase_shift_rad
    if not isinstance(bfactor, list) and type(bfactor) is not np.ndarray:
        cs_data['ctf/bfactor'] = [bfactor] * data_len
    else:
        cs_data['ctf/bfactor'] = bfactor
    return cs_data



# Copyright (C) 2017 Daniel Asarnow
# University of California, San Francisco
#
# Library functions for volume data.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
import numpy.ma as ma
from scipy.ndimage import map_coordinates
from pyfftw.interfaces.numpy_fft import rfftn
from .vop_numba import fill_ft


def ismask(vol):
    """
    Even with a soft edge, a mask will have very few unique values (unless it's already been resampled).
    The 1D slice below treats just the central XY section for speed. Real maps have ~20,000 unique values here.
    """
    return np.unique(vol[vol.shape[2] // 2::vol.shape[2]]).size < 100


def resample_volume(vol, r=None, t=None, ori=None, order=3, compat="mrc2014",
                    indexing="ij", invert=False, scale=None, output_shape=None):
    if r is None and t is None and scale is None and (output_shape is None or np.array_equal(output_shape, vol.shape)):
        return vol.copy()

    if output_shape is None:
        output_shape = np.array(vol.shape)
    elif np.isscalar(output_shape):
        output_shape = np.array((output_shape, output_shape, output_shape))

    x, y, z = np.meshgrid(*[np.arange(-c, c) for c in output_shape // 2], indexing=indexing)
    xyz = np.vstack([x.reshape(-1), y.reshape(-1), z.reshape(-1), np.ones(x.size)])

    if ori is not None:
        xyz -= ori[:, None]

    if r is None:
        r = np.eye(3)

    th = np.eye(4)
    if t is None and r.shape[1] == 4:
        t = np.squeeze(r[:, 3])
    elif t is not None:
        th[:3, 3] = t

    rh = np.eye(4)
    rh[:3, :3] = r[:3, :3].T

    if scale is not None:
        rh[:3, :3] /= scale

    center = np.array(vol.shape) // 2

    if invert:
        th[:3, 3] = -th[:3, 3]
        rh[:3, :3] = rh[:3:, :3].T
        xyz = rh.dot(th.dot(xyz))[:3, :] + center[:, None]
    else:
        xyz = th.dot(rh.dot(xyz))[:3, :] + center[:, None]

    xyz = np.array([arr.reshape(output_shape) for arr in xyz])

    if "relion" in compat.lower() or "xmipp" in compat.lower():
        xyz = xyz[::-1]

    newvol = map_coordinates(vol, xyz, order=order)
    return newvol


def grid_correct(vol, pfac=2, order=1):
    n = vol.shape[0]
    nhalf = n // 2
    x, y, z = np.meshgrid(*[np.arange(-nhalf, nhalf)] * 3, indexing="xy")
    r = np.sqrt(x**2 + y**2 + z**2, dtype=vol.dtype) / (n * pfac)
    with np.errstate(divide="ignore", invalid="ignore"):
        sinc = np.sin(np.pi * r) / (np.pi * r)  # Results in 1 NaN in the center.
    sinc[nhalf, nhalf, nhalf] = 1.
    if order == 0:
       cordata = vol / sinc
    elif order == 1:
       cordata = vol / sinc**2
    else:
        raise NotImplementedError("Only nearest-neighbor and trilinear grid corrections are available")
    return cordata


def interpolate_slice(f3d, rot, pfac=2, size=None):
    nhalf = f3d.shape[0] / 2
    if size is None:
        phalf = nhalf
    else:
        phalf = size / 2
    qot = rot * pfac  # Scaling!
    px, py, pz = np.meshgrid(np.arange(-phalf, phalf), np.arange(-phalf, phalf), 0)
    pr = np.sqrt(px ** 2 + py ** 2 + pz ** 2)
    pcoords = np.vstack([px.reshape(-1), py.reshape(-1), pz.reshape(-1)])
    mcoords = qot.T.dot(pcoords)
    mcoords = mcoords[:, pr.reshape(-1) < nhalf]
    pvals = map_coordinates(np.real(f3d), mcoords, order=1, mode="wrap") + \
             1j * map_coordinates(np.imag(f3d), mcoords, order=1, mode="wrap")
    pslice = np.zeros(pr.shape, dtype=np.complex)
    pslice[pr < nhalf] = pvals
    return pslice


def vol_ft(vol, pfac=2, threads=1, normfft=1):
    """ Returns a centered, Nyquist-limited, zero-padded, interpolation-ready 3D Fourier transform.
    :param vol: Volume to be Fourier transformed.
    :param pfac: Size factor for zero-padding.
    :param threads: Number of threads for pyFFTW.
    :param normfft: Normalization constant for Fourier transform.
    """
    vol = grid_correct(vol, pfac=pfac, order=1)
    padvol = np.pad(vol, int((vol.shape[0] * pfac - vol.shape[0]) // 2), "constant")
    ft = rfftn(np.fft.ifftshift(padvol), padvol.shape, threads=threads)
    ftc = np.zeros((ft.shape[0] + 3, ft.shape[1] + 3, ft.shape[2]), dtype=ft.dtype)
    fill_ft(ft, ftc, vol.shape[0], normfft=normfft)
    return ftc


def normalize(vol, ref=None, return_stats=False, rmask=1.0):
    volm = vol.view(ma.MaskedArray)
    sz = volm.shape[0]
    half = sz / 2
    rng = np.arange(-half, half)
    x, y, z = np.meshgrid(rng, rng, rng)
    r2 = x**2 + y**2 + z**2
    rmask = (half * rmask) if rmask <= 1.0 else rmask
    mask = r2 > rmask**2
    volm.mask = mask
    if ref is not None:
        ref = ref.view(ma.MaskedArray)
        ref.mask = mask
        sigma = np.std(ref)
        mu = np.mean(ref)
    else:
        sigma = np.std(volm)
        mu = np.mean(volm)
    if return_stats:
        return (vol - mu) / sigma, mu, sigma
    return (vol - mu) / sigma


def label_to_axis(label):
    label = label.lower()
    if label == "z":
        return 2
    if label == "y":
        return 1
    if label == "x":
        return 0
    return None




# -*- coding: utf-8 -*-
# Copyright (C) 2016 Eugene Palovcak and Daniel Asarnow
# University of Calfornia, San Francisco
#
# Library functions for volume data.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
import os


MODE = {0: np.dtype(np.int8), 1: np.dtype(np.int16), 2: np.dtype(np.float32),
        6: np.dtype(np.uint16), 12: np.dtype(np.float16),
        np.dtype(np.int8): 0, np.dtype(np.int16): 1, np.dtype(np.float32): 2,
        np.dtype(np.uint16): 6, np.dtype(np.float16): 12}
HEADER_LEN = int(1024)  # Bytes.


def mrc_header(shape, dtype=np.float32, psz=1.0):
    header = np.zeros(HEADER_LEN // np.dtype(np.int32).itemsize, dtype=np.int32)
    header_f = header.view(np.float32)
    header[:3] = shape
    if np.dtype(dtype) not in MODE:
        raise ValueError("Invalid dtype for MRC")
    header[3] = MODE[np.dtype(dtype)]
    header[7:10] = header[:3]  # mx, my, mz (grid size)
    header_f[10:13] = psz * header[:3]  # xlen, ylen, zlen
    header_f[13:16] = 90.0  # CELLB
    header[16:19] = 1, 2, 3  # Axis order.
    header_f[19:22] = 1, 0, -1  # Convention for unreliable  values.
    # header[26] = 1329812045  # "MRCO" chars.
    header[27] = 20140  # Version 2014-0.
    header_f[49:52] = 0, 0, 0  # Default origin.
    header[52] = 542130509  # "MAP " chars.
    header[53] = 17476  # 0x00004444 for little-endian.
    header_f[54] = -1  # Convention for unreliable RMS value.
    return header


def mrc_header_complete(data, psz=1.0, origin=None):
    header = mrc_header(data.shape, data.dtype, psz=psz)
    header_f = header.view(np.float32)
    header_f[19:22] = [data.min(), data.max(), data.mean()]
    header_f[54] = np.sqrt(np.mean(data**2))
    if origin is None:
        header_f[49:52] = (0, 0, 0)
    elif origin == "center":
        header_f[49:52] = psz * header[:3] / 2
    else:
        header_f[49:52] = origin
    return header


def _read_header(f):
    hdr = {}
    pos = f.tell()
    f.seek(0)
    header = np.fromfile(f, dtype=np.int32, count=256)
    f.seek(pos)
    header_f = header.view(np.float32)
    [hdr['nx'], hdr['ny'], hdr['nz'], hdr['datatype']] = header[:4]
    [hdr['xlen'], hdr['ylen'], hdr['zlen']] = header_f[10:13]
    if hdr['xlen'] == hdr['ylen'] == hdr['zlen'] == 0:
        hdr['xlen'] = hdr['nx']
        hdr['ylen'] = hdr['ny']
        hdr['zlen'] = hdr['nz']
    return hdr


def read_header(fname):
    with open(fname) as f:
        hdr = _read_header(f)
    return hdr


def read(fname, inc_header=False, compat="mrc2014"):
    if "relion" in compat.lower() or "xmipp" in compat.lower():
        order = "C"
    else:
        order = "F"  # Read x, y, z -> data[i, j, k] == data[i][j][k]
    data = None
    with open(fname) as f:
        hdr = _read_header(f)
        nx = hdr['nx']
        ny = hdr['ny']
        nz = hdr['nz']
        datatype = hdr['datatype']
        f.seek(HEADER_LEN)  # seek to start of data
        if nz == 1:
            shape = (nx, ny)
        else:
            shape = (nx, ny, nz)
        if datatype in MODE:
            dtype = MODE[datatype]
        else:
            raise IOError("Unknown MRC data type %d" % datatype)
        data = np.reshape(np.fromfile(f, dtype=dtype, count=nx * ny * nz), shape, order=order)
    if inc_header:
        return data, hdr
    else:
        return data


def write(fname, data, psz=1, origin=None, fast=False):
    """
    Write a MRC file. Fortran axes order is assumed.
    :param fname: Destination path.
    :param data: Array to write.
    :param psz: Pixel size in Å for MRC header.
    :param origin: Coordinate of origin voxel.
    :param fast: Skip computing density statistics in header. Default is False.
    """
    data = np.atleast_3d(data)
    if fast:
        header = mrc_header(data.shape, dtype=data.dtype, psz=psz)
    else:
        header = mrc_header_complete(data, psz=psz, origin=origin)
    with open(fname, 'wb') as f:
        f.write(header.tobytes())
        f.write(np.require(data, dtype=np.float32).tobytes(order="F"))


def append(fname, data):
    data = np.atleast_3d(data)
    with open(fname, 'r+b') as f:
        nx, ny, nz = np.fromfile(f, dtype=np.int32, count=3)
        f.seek(36)  # First byte of zlen.
        zlen = np.fromfile(f, dtype=np.float32, count=1)
        if data.shape[0] != nx or data.shape[1] != ny:
            raise ValueError("Data has different shape than destination file")
        f.seek(0, os.SEEK_END)
        f.write(np.require(data, dtype=np.float32).tobytes(order="F"))
        # Update header after new data is written.
        apix = zlen / nz
        nz += data.shape[2]
        zlen += apix * data.shape[2]
        f.seek(8)
        nz.astype(np.int32).tofile(f)
        f.seek(36)
        zlen.astype(np.float32).tofile(f)


def write_imgs(fname, idx, data):
    with open(fname, 'r+b') as f:
        nx, ny, nz, dtype = np.fromfile(f, dtype=np.int32, count=4)
        if dtype in MODE:
            dtype = MODE[dtype]
        else:
            raise ValueError("Invalid dtype for MRC")
        if not np.can_cast(data.dtype, dtype):
            raise ValueError("Can't cast %s to %s" % (data.dtype, dtype))
        if data.shape[2] > nz:
            raise ValueError("Data has more z-slices than destination file")
        if data.shape[0] != nx or data.shape[1] != ny:
            raise ValueError("Data has different shape than destination file")
        f.seek(HEADER_LEN + idx * nx * ny * dtype.itemsize)
        f.write(np.require(data, dtype=dtype).tobytes(order="F"))


def read_imgs(fname, idx, num=1, compat="mrc2014"):
    if "relion" in compat or "xmipp" in compat:
        order = "C"
    else:
        order = "F"
    with open(fname) as f:
        nx, ny, nz, datatype = np.fromfile(f, dtype=np.int32, count=4)
        assert (idx < nz)
        if num < 0:
            num = nz - idx
        assert (idx + num <= nz)
        assert (num != 0)
        if num == 1:
            shape = (nx, ny)
        else:
            shape = (nx, ny, num)
        if datatype in MODE:
            dtype = MODE[datatype]
        else:
            raise IOError("Unknown MRC data type %d" % datatype)
        f.seek(HEADER_LEN + idx * dtype.itemsize * nx * ny)
        return np.reshape(np.fromfile(f, dtype=dtype, count=nx * ny * num),
                          shape, order=order)


def read_zslices(fname):
    with ZSliceReader(fname) as zsr:
        for i in range(zsr.nz):
            yield zsr.read(i)


class ZSliceReader:
    def __init__(self, fname):
        self.path = fname
        self.f = open(self.path)
        self.nx, self.ny, self.nz, datatype = np.fromfile(self.f, dtype=np.int32, count=4)
        self.shape = (self.nx, self.ny)
        self.size = self.nx * self.ny
        if datatype in MODE:
            self.dtype = MODE[datatype]
        else:
            raise IOError("Unknown MRC data type %d" % datatype)
        self.i = 0

    def read(self, i):
        self.i = i
        if self.i >= self.nz:
            raise IOError("Index %d out of bounds for stack of size %d" % (i, self.nz))
        self.f.seek(HEADER_LEN + self.i * self.dtype.itemsize * self.size)
        # Populate slice C order so that X is fastest axis i.e. j / columns.
        return np.reshape(
            np.fromfile(self.f, dtype=self.dtype, count=self.size), self.shape)

    def close(self):
        self.f.close()

    def __iter__(self):
        self.i = 0
        return self

    def next(self):
        try:
            item = self.read(self.i)
        except IOError:
            raise StopIteration
        self.i += 1
        return item

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class ZSliceWriter:
    def __init__(self, fname, shape=None, dtype=np.float32, psz=1.0, mode="w"):
        self.path = fname
        self.shape = None
        self.size = None
        self.psz = psz
        self.dtype = None
        self.f = None
        self.i = 0
        if shape is not None:
            self.set_shape(shape)
        if dtype is not None:
            self.set_dtype(dtype)
        if "a" in mode:
            hdr = read_header(self.path)
            self.f = open(self.path, 'ab')
            self.psz = hdr["xlen"] / hdr["nx"]
            self.set_shape((hdr["nx"], hdr["ny"]))
            if hdr["datatype"] in MODE:
                self.set_dtype(hdr["datatype"])
            else:
                raise IOError("Unknown MRC data type %d" % hdr['datatype'])
            self.f.seek(0, os.SEEK_END)
        else:
            self.f = open(self.path, 'wb')
            # self.f.seek(HEADER_LEN)  # Results in a sparse file?
            self.f.write(b'\x00' * HEADER_LEN)

    def set_dtype(self, dtype):
        if np.dtype(dtype) not in MODE:
            raise ValueError("Unknown MRC data type %d" % dtype)
        self.dtype = np.dtype(dtype)

    def set_shape(self, shape):
        if len(shape) == 1:
            self.shape = (1, shape[0])
        elif len(shape) == 2:
            self.shape = (shape[0], shape[1])
        elif len(shape) == 3:
            self.shape = (shape[1], shape[2])
        else:
            raise ValueError("Number of dimensions must be 1, 2, or 3")
        self.size = np.prod(self.shape)

    def write(self, arr):
        """
        Write one or more z-slices to ZSliceWriter's underlying File object.
        If ZSliceWriter was constructed without a shape, then the first two
        dimensions of the array define the shape (and slice size).
        Subsequent arrays can be of any shape, as long as their size is an
        integer multiple of the slice size.
        :param arr: A numpy array.
        """
        if self.i == 0:
            if self.shape is None:
                self.set_shape(arr.shape)
            if self.dtype is None:
                self.set_dtype(arr.dtype)
        assert np.can_cast(arr.dtype, self.dtype, casting="same_kind")
        assert arr.size % self.size == 0
        # self.f.seek(HEADER_LEN + self.i * self.dtype.itemsize * arr.size)
        self.f.write(np.require(arr, dtype=self.dtype).tobytes())
        self.i += arr.size / self.size

    def close(self):
        header = mrc_header(shape=(self.shape[1], self.shape[0], self.i),
                            dtype=self.dtype, psz=self.psz)
        self.f.seek(0)
        self.f.write(header.tobytes())
        self.f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


# HEADER FORMAT
# 0      (0,4)      NX  number of columns (fastest changing in map)
# 1      (4,8)      NY  number of rows
# 2      (8,12)     NZ  number of sections (slowest changing in map)
# 3      (12,16)    MODE  data type:
#                       0   image: signed 8-bit bytes range -128 to 127
#                       1   image: 16-bit halfwords
#                       2   image: 32-bit reals
#                       3   transform: complex 16-bit integers
#                       4   transform: complex 32-bit reals
# 4      (16,20)    NXSTART number of first column in map
# 5      (20,24)    NYSTART number of first row in map
# 6      (24,28)    NZSTART number of first section in map
# 7      (28,32)    MX      number of intervals along X
# 8      (32,36)    MY      number of intervals along Y
# 9      (36,40)    MZ      number of intervals along Z
# 10-13  (40,52)    CELLA   cell dimensions in angstroms
# 13-16  (52,64)    CELLB   cell angles in degrees
# 16     (64,68)    MAPC    axis corresp to cols (1,2,3 for X,Y,Z)
# 17     (68,72)    MAPR    axis corresp to rows (1,2,3 for X,Y,Z)
# 18     (72,76)    MAPS    axis corresp to sections (1,2,3 for X,Y,Z)
# 19     (76,80)    DMIN    minimum density value
# 20     (80,84)    DMAX    maximum density value
# 21     (84,88)    DMEAN   mean density value
# 22     (88,92)    ISPG    space group number, 0 for images or 1 for volumes
# 23     (92,96)    NSYMBT  number of bytes in extended header
# 24-49  (96,196)   EXTRA   extra space used for anything
#           26  (104)   EXTTYP      extended header type("MRCO" for MRC)
#           27  (108)   NVERSION    MRC format version (20140)
# 49-52  (196,208)  ORIGIN  origin in X,Y,Z used for transforms
# 52     (208,212)  MAP     character string 'MAP ' to identify file type
# 53     (212,216)  MACHST  machine stamp
# 54     (216,220)  RMS     rms deviation of map from mean density
# 55     (220,224)  NLABL   number of labels being used
# 56-256 (224,1024) LABEL(80,10)    10 80-character text labels



# Copyright (C) 2017-2018 Daniel Asarnow
# University of California, San Francisco
#
# Library functions for volume data.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numba
import numpy as np


@numba.jit(cache=True, nopython=True, nogil=True)
def fill_ft(ft, ftc, rmax, normfft=1):
    rmax2 = rmax ** 2
    for k in range(ft.shape[0]):
        kp = k if k < ft.shape[2] else k - ft.shape[0]
        for i in range(ft.shape[1]):
            ip = i if i < ft.shape[2] else i - ft.shape[1]
            for j in range(ft.shape[2]):
                jp = j
                r2 = ip**2 + jp**2 + kp**2
                if r2 <= rmax2:
                    ftc[kp + ftc.shape[0]//2, ip + ftc.shape[1]//2, jp] = ft[k, i, j] * normfft


@numba.jit(cache=False, nopython=True, nogil=True)
def interpolate_slice_numba(f3d, rot, pfac=2, size=None):
    linterp = lambda a, l, h: l + (h - l) * a
    ori = f3d.shape[0] // 2 - 1
    n = (f3d.shape[0] - 3) // pfac
    nhalf = n // 2
    if size is None:
        size = n
    phalf = size // 2
    rmax = min(nhalf, (phalf+1) - 1)
    rmax2 = rmax**2
    qot = rot.T * pfac  # Scaling!
    f2d = np.zeros((np.int64(size), np.int64(phalf + 1)), dtype=np.complex64)

    for i in range(size):
        
        if i <= rmax:
            yp = i
        elif i >= size - rmax:
            yp = i - size
        else:
            continue

        yp2 = yp ** 2

        for j in range(rmax + 1):
            xp = j

            if xp**2 + yp2 > rmax2:
                continue

            x = qot[0,0] * xp + qot[0,1] * yp  # Implicit z = 0.
            y = qot[1,0] * xp + qot[1,1] * yp
            z = qot[2,0] * xp + qot[2,1] * yp

            if x < 0:
                x = -x
                y = -y
                z = -z
                negx = True
            else:
                negx = False

            x0 = int(np.floor(x))
            y0 = int(np.floor(y))
            z0 = int(np.floor(z))

            ax = x - x0
            ay = y - y0
            az = z - z0

            y0 += ori + 1
            z0 += ori + 1

            x1 = x0 + 1
            y1 = y0 + 1
            z1 = z0 + 1

            f2d[i, j] = linterp(az,
                    linterp(ay,
                            linterp(ax, f3d[z0,y0,x0], f3d[z0,y0,x1]),
                            linterp(ax, f3d[z0,y1,x0], f3d[z0,y1,x1])),
                    linterp(ay,
                            linterp(ax, f3d[z1,y0,x0], f3d[z1,y0,x1]),
                            linterp(ax, f3d[z1,y1,x0], f3d[z1,y1,x1])))

            if negx:
                f2d[i, j] = np.conj(f2d[i, j])
    return f2d


def accumulate_slice_nb(f3d, f2d, rot, pfac=2):
    pass



import numpy as np
import torch
from torch.fft import fftshift, fft2


def fft2_center(img, tensor=False):
    if tensor:
        return fftshift(fft2(fftshift(img, dim=(-1, -2))), dim=(-1, -2))
    else:
        return np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(img, axes=(-1, -2))), axes=(-1, -2))


def fftn_center(img):
    return np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(img)))


def ifftn_center(vol):
    vol = np.fft.ifftshift(vol)
    vol = np.fft.ifftn(vol)
    vol = np.fft.fftshift(vol)
    return vol


def ht2_center(img, tensor=False):
    f = fft2_center(img, tensor=tensor)
    return f.real - f.imag


def htn_center(img):
    f = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(img)))
    return f.real - f.imag


def iht2_center(img):
    img = fft2_center(img)
    img /= (img.shape[-1] * img.shape[-2])
    return img.real - img.imag


def ihtn_center(vol):
    vol = np.fft.ifftshift(vol)
    vol = np.fft.fftn(vol)
    vol = np.fft.fftshift(vol)
    vol /= np.prod(vol.shape)
    return vol.real - vol.imag


def symmetrize_ht(ht, pre_allocated=False):
    if pre_allocated:
        resolution = ht.shape[-1] - 1
        sym_ht = ht
    else:
        if len(ht.shape) == 2:
            ht = ht.reshape(1, *ht.shape)
        assert len(ht.shape) == 3
        resolution = ht.shape[-1]
        batch_size = ht.shape[0]
        sym_ht = np.empty((batch_size, resolution + 1, resolution + 1), dtype=ht.dtype)
        sym_ht[:, 0:-1, 0:-1] = ht
    assert resolution % 2 == 0
    sym_ht[:, -1, :] = sym_ht[:, 0]  # last row is the first row
    sym_ht[:, :, -1] = sym_ht[:, :, 0]  # last col is the first col
    sym_ht[:, -1, -1] = sym_ht[:, 0, 0]
    if len(sym_ht) == 1:
        sym_ht = sym_ht[0]
    return sym_ht


def symmetrize_ht_torch(ht):
    if ht.ndim == 2:
        ht = ht[np.newaxis, ...]
    assert ht.ndim == 3
    n = ht.shape[0]

    D = ht.shape[-1]
    sym_ht = torch.empty((n, D + 1, D + 1), dtype=ht.dtype, device=ht.device)
    sym_ht[:, 0:-1, 0:-1] = ht

    assert D % 2 == 0
    sym_ht[:, -1, :] = sym_ht[:, 0, :]  # last row is the first row
    sym_ht[:, :, -1] = sym_ht[:, :, 0]  # last col is the first col
    sym_ht[:, -1, -1] = sym_ht[:, 0, 0]  # last corner is first corner

    if n == 1:
        sym_ht = sym_ht[0, ...]

    return sym_ht


# Copyright (C) 2017-2022 Daniel Asarnow
# University of California, San Francisco
#
# Handles metadata from cryoSPARC 2.0 and later.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
import numpy as np
import os.path
import pandas as pd
from .. import geom
from .. import star
from .. import util


micrograph = {u'uid': star.UCSF.UID,
              u'micrograph_blob/path': star.Relion.MICROGRAPH_NAME,
              u'micrograph_blob/psize_A': star.Relion.MICROGRAPHPIXELSIZE,
              u'mscope_params/accel_kv': star.Relion.VOLTAGE,
              u'mscope_params/cs_mm': star.Relion.CS,
              u'ctf/accel_kv': star.Relion.VOLTAGE,
              u'ctf/amp_contrast': star.Relion.AC,
              u'ctf/cs_mm': star.Relion.CS,
              u'ctf/df1_A': star.Relion.DEFOCUSU,
              u'ctf/df2_A': star.Relion.DEFOCUSV,
              u'ctf/df_angle_rad': star.Relion.DEFOCUSANGLE,
              u'ctf/phase_shift_rad': star.Relion.PHASESHIFT,
              u'ctf/cross_corr_ctffind4': star.Relion.CTFFIGUREOFMERIT,
              u'ctf/ctf_fit_to_A': star.Relion.CTFMAXRESOLUTION}
general = {u'uid': star.UCSF.UID,
           u'ctf/accel_kv': star.Relion.VOLTAGE,
           u'blob/psize_A': star.Relion.IMAGEPIXELSIZE,
           u'ctf/ac': star.Relion.AC,
           u'ctf/amp_contrast': star.Relion.AC,
           u'ctf/cs_mm': star.Relion.CS,
           u'ctf/df1_A': star.Relion.DEFOCUSU,
           u'ctf/df2_A': star.Relion.DEFOCUSV,
           u'ctf/df_angle_rad': star.Relion.DEFOCUSANGLE,
           u'ctf/phase_shift_rad': star.Relion.PHASESHIFT,
           u'ctf/cross_corr_ctffind4': star.Relion.CTFFIGUREOFMERIT,
           u'ctf/ctf_fit_to_A': star.Relion.CTFMAXRESOLUTION,
           u'ctf/bfactor': star.Relion.CTFBFACTOR,
           u'ctf/exp_group_id': star.Relion.OPTICSGROUP,
           u'blob/path': star.UCSF.IMAGE_PATH,
           u'blob/idx': star.UCSF.IMAGE_INDEX,
           u'location/center_x_frac': None,
           u'location/center_y_frac': None,
           u'location/micrograph_path': star.Relion.MICROGRAPH_NAME,
           u'location/micrograph_shape': None,
           u'filament/filament_uid': star.Relion.HELICALTUBEID,
           u'filament/filament_pose': None}
movie = {u'movie_blob/path': star.Relion.MICROGRAPHMOVIE_NAME,
         u'movie_blob/psize_A': star.Relion.MICROGRAPHORIGINALPIXELSIZE,
         u'gain_ref_blob/path': star.Relion.MICROGRAPHGAIN_NAME,
         u'rigid_motion/frame_start': star.Relion.MICROGRAPHSTARTFRAME,
         u'rigid_motion/frame_end': star.Relion.MICROGRAPHENDFRAME,
         u'mscope_params/total_dose_e_per_A2': star.Relion.MICROGRAPHDOSERATE}


def cryosparc_2_cs_particle_locations(cs, df=None, swapxy=True, invertx=False, inverty=True):
    log = logging.getLogger('root')
    if df is None:
        df = pd.DataFrame()
    if u'location/center_x_frac' in cs.dtype.names:
        log.debug("Converting normalized particle coordinates to absolute")
        df[star.Relion.COORDX] = cs[u'location/center_x_frac']
        df[star.Relion.COORDY] = cs[u'location/center_y_frac']
        # df[star.Relion.MICROGRAPH_NAME] = cs[u'location/micrograph_path']
        if invertx:
            # Might rarely be needed, if your K3 images are "tall" in SerialEM.
            df[star.Relion.COORDX] = 1 - df[star.Relion.COORDX]
        if inverty:
            # cryoSPARC coordinates have origin in "bottom left" so inverting Y is default for Relion correctness
            # (and therefore also for Import Particles). However, cryoSPARC Patch Motion flips images physically
            # vs. SerialEM, Motioncor2 doesn't, so "inverting twice" (not inverting) is required if switching.
            df[star.Relion.COORDY] = 1 - df[star.Relion.COORDY]
        if swapxy:
            # In cryoSPARC, fast axis is long axis of K3, 'location/micrograph_shape' is [y, x].
            # In Relion and numpy (e.g. pyem.mrc), the fast axis is the short axis of K3, shape is (x, y).
            # cryoSPARC import particles correctly imports *Relion convention* coordinates, which we also want.
            # Default behavior is now to always swap.
            df[star.Relion.COORDS] = np.round(df[star.Relion.COORDS] *
                                              cs['location/micrograph_shape'][:, ::-1]).astype(int)
        else:
            df[star.Relion.COORDS] = np.round(df[star.Relion.COORDS] * cs['location/micrograph_shape']).astype(int)
        log.info("Converted particle coordinates from normalized to absolute")
    return df


def cryosparc_2_cs_ctf_parameters(cs, df=None):
    log = logging.getLogger('root')
    if df is None:
        df = pd.DataFrame()
    if 'ctf/tilt_A' in cs.dtype.names:
        log.debug("Recovering beam tilt and converting to mrad")
        df[star.Relion.BEAMTILTX] = np.arcsin(cs['ctf/tilt_A'][:, 0] / cs['ctf/cs_mm'] * 1e-7) * 1e3
        df[star.Relion.BEAMTILTY] = np.arcsin(cs['ctf/tilt_A'][:, 1] / cs['ctf/cs_mm'] * 1e-7) * 1e3
    if 'ctf/shift_A' in cs.dtype.names:
        pass
    if 'ctf/trefoil_A' in cs.dtype.names:
        pass
        # df[star.Relion.ODDZERNIKE] = cs['ctf/trefoil_A']
    if 'ctf/tetrafoil_A' in cs.dtype.names:
        pass
        # df[star.Relion.EVENZERNIKE] = cs['ctf/tetra_A']
    if 'ctf/anisomag' in cs.dtype.names:
        df[star.Relion.MAGMAT00] = cs['ctf/anisomag'][:, 0]
        df[star.Relion.MAGMAT01] = cs['ctf/anisomag'][:, 1]
        df[star.Relion.MAGMAT10] = cs['ctf/anisomag'][:, 2]
        df[star.Relion.MAGMAT11] = cs['ctf/anisomag'][:, 3]
    return df


def cryosparc_2_cs_model_parameters(cs, df=None, minphic=0):
    model = {u'split': star.Relion.RANDOMSUBSET,
             u'shift': star.Relion.ORIGINS,
             u'pose': star.Relion.ANGLES,
             u'error': None,
             u'error_min': None,
             u'resid_pow': None,
             u'slice_pow': None,
             u'image_pow': None,
             u'cross_cor': None,
             u'alpha': None,
             u'weight': None,
             u'pose_ess': None,
             u'shift_ess': None,
             u'class_posterior': None,
             u'class': star.Relion.CLASS,
             u'class_ess': None}
    log = logging.getLogger('root')
    if df is None:
        df = pd.DataFrame()
    phic_names = [n for n in cs.dtype.names if "class_posterior" in n]
    if u'alignments3D/class_posterior' in cs.dtype.names:
        log.info("Assigning pose from single 3D refinement")
        for k in model:
            if model[k] is not None:
                name = u'alignments3D/' + k
                df[model[k]] = pd.DataFrame(cs[name])
    elif len(phic_names) > 1:
        log.info("Assigning pose from most likely 3D classes")
        phic = np.array([cs[p] for p in phic_names if u'alignments2D' not in p])
        cls = np.argmax(phic, axis=0)
        cls_prob = phic[cls, range(cls.shape[0])]
        for k in model:
            if model[k] is not None:
                names = [n for n in cs.dtype.names if n.endswith(k)]
                df[model[k]] = pd.DataFrame(np.array(
                    [cs[names[c]][i] for i, c in enumerate(cls)]))
        if minphic > 0:
            df.drop(df.loc[cls_prob < minphic].index, inplace=True)
    elif u'alignments2D/class_posterior' in cs.dtype.names:
        log.info("Assigning pose from 2D classes")
        log.info("Assigning skew angle from 2D classification")
        model["pose"] = star.Relion.ANGLEPSI
        for k in model:
            if model[k] is not None:
                name = "alignments2D/" + k
                df[model[k]] = pd.DataFrame(cs[name])
    else:
        log.info("Particle poses not found")
    return df


def cryosparc_2_cs_array_parameters(cs, df=None):
    log = logging.getLogger('root')
    if df is None:
        df = pd.DataFrame()
    if "blob/shape" in cs.dtype.names:
        log.info("Copying image size")
        df[star.Relion.IMAGESIZE] = cs["blob/shape"][:, 0]
    elif "movie_blob/shape" in cs.dtype.names:
        log.info("Copying movie size")
        df[star.Relion.IMAGESIZEX] = cs["movie_blob/shape"][:, 2]
        df[star.Relion.IMAGESIZEY] = cs["movie_blob/shape"][:, 1]
        df[star.Relion.IMAGESIZEZ] = cs["movie_blob/shape"][:, 0]
    elif "micrograph_blob/shape" in cs.dtype.names:
        log.info("Copying micrograph size")
        df[star.Relion.IMAGESIZEX] = cs["micrograph_blob/shape"][:, 1]
        df[star.Relion.IMAGESIZEY] = cs["micrograph_blob/shape"][:, 0]
    return df


def cryosparc_2_cs_filament_parameters(cs, df=None):
    log = logging.getLogger('root')
    if df is None:
        df = pd.DataFrame()
    if 'filament/filament_pose' in cs.dtype.names:
        log.info('Copying filament pose')
        df[star.Relion.ANGLEPSIPRIOR] = np.rad2deg(-cs['filament/filament_pose'] + np.pi / 2)
    return df


def cryosparc_2_cs_movie_parameters(cs, passthroughs=None, trajdir=".", path=None):
    log = logging.getLogger('root')
    log.info("Creating movie data_general tables")
    data_general = util.dataframe_from_records_mapped(cs, {**movie, **micrograph, **general})
    data_general = cryosparc_2_cs_array_parameters(cs, data_general)
    if passthroughs is not None:
        for passthrough in passthroughs:
            log.info("Reading auxiliary file %s" % passthrough)
            pt = np.load(passthrough)
            ptdf = util.dataframe_from_records_mapped(pt, {**movie, **micrograph, **general})
            ptdf = cryosparc_2_cs_array_parameters(cs, ptdf)
            key = star.UCSF.UID
            fields = [c for c in ptdf.columns if c not in data_general.columns]
            data_general = star.smart_merge(data_general, ptdf, fields=fields, key=key)
    data_general[star.Relion.MOTIONMODELVERSION] = 0
    data_general[star.Relion.MICROGRAPHBINNING] = \
        data_general[star.Relion.MICROGRAPHPIXELSIZE] / data_general[star.Relion.MICROGRAPHORIGINALPIXELSIZE]
    data_general[star.Relion.MICROGRAPHDOSERATE] /= data_general[star.Relion.IMAGESIZEZ]
    data_general[star.Relion.MICROGRAPHPREEXPOSURE] = \
        data_general[star.Relion.MICROGRAPHDOSERATE] * data_general[star.Relion.MICROGRAPHSTARTFRAME]
    data_general[star.Relion.MICROGRAPHSTARTFRAME] += 1
    data_general = star.decode_byte_strings(data_general, fmt='UTF-8', inplace=True)
    if path is None:
        data_general[star.Relion.MICROGRAPHMOVIE_NAME] = data_general[star.Relion.MICROGRAPHMOVIE_NAME].apply(
            lambda x: os.path.join(trajdir, x))
        if star.Relion.MICROGRAPHGAIN_NAME in data_general:
            data_general[star.Relion.MICROGRAPHGAIN_NAME] = data_general[star.Relion.MICROGRAPHGAIN_NAME].apply(
                lambda x: os.path.join(trajdir, x))
        else:
            log.warning("No gain reference found")
        data_general[star.Relion.MICROGRAPH_NAME] = data_general[star.Relion.MICROGRAPH_NAME].apply(
            lambda x: os.path.join(trajdir, x))
    else:
        data_general[star.Relion.MICROGRAPHMOVIE_NAME] = data_general[star.Relion.MICROGRAPHMOVIE_NAME].apply(
            lambda x: os.path.join(path, os.path.basename(x)))
        if star.Relion.MICROGRAPHGAIN_NAME in data_general:
            data_general[star.Relion.MICROGRAPHGAIN_NAME] = data_general[star.Relion.MICROGRAPHGAIN_NAME].apply(
                lambda x: os.path.join(path, os.path.basename(x)))
        else:
            log.warning("No gain reference found")
        data_general[star.Relion.MICROGRAPH_NAME] = data_general[star.Relion.MICROGRAPH_NAME].apply(
            lambda x: os.path.join(path, os.path.basename(x)))
    return data_general


def cryosparc_2_cs_motion_parameters(cs, data, trajdir="."):
    log = logging.getLogger('root')
    log.info("Reading movie trajectory files")
    for i in range(cs.shape[0]):
        trajfile = cs['rigid_motion/path'][i].decode('UTF-8')
        trajfile = os.path.join(trajdir, trajfile)
        traj = np.load(trajfile).reshape((-1, 2))
        zsf = 0  # Or cs['rigid_motion/zero_shift_frame'][i] ?
        traj[:, 0] = -traj[:, 0] + traj[zsf, 0]  # Invert vectors and re-center to frame 0.
        traj[:, 1] = traj[:, 1] - traj[zsf, 1]  # Invert Y here.
        traj /= cs['rigid_motion/psize_A'][i]  # Looks right.
        log.debug("%s: %d-%d, (%d x %d)" %
                  (trajfile, cs['rigid_motion/frame_start'][i], cs['rigid_motion/frame_end'][i],
                  traj.shape[0], traj.shape[1]))
        d = {star.Relion.MICROGRAPHFRAMENUMBER: np.arange(cs['rigid_motion/frame_start'][i] + 1,
                                                          cs['rigid_motion/frame_end'][i] + 1),
             star.Relion.MICROGRAPHSHIFTX: traj[:, 0],
             star.Relion.MICROGRAPHSHIFTY: traj[:, 1]}
        try:
            data_shift = pd.DataFrame(d)
            mic = {star.Relion.GENERALDATA: data.iloc[i], star.Relion.GLOBALSHIFTDATA: data_shift}
        except ValueError:
            log.debug("Couldn't convert %s, skipping" % trajfile)
            continue
        yield mic


def parse_cryosparc_2_cs(csfile, passthroughs=None, minphic=0, boxsize=None,
                         swapxy=False, invertx=False, inverty=False):

    log = logging.getLogger('root')
    log.debug("Reading primary file")
    cs = csfile if type(csfile) is np.ndarray else np.load(csfile)
    df = util.dataframe_from_records_mapped(cs, general)
    df = cryosparc_2_cs_particle_locations(cs, df, swapxy=swapxy, invertx=invertx, inverty=inverty)
    df = cryosparc_2_cs_model_parameters(cs, df, minphic=minphic)
    df = cryosparc_2_cs_array_parameters(cs, df)
    df = cryosparc_2_cs_filament_parameters(cs, df)
    if passthroughs is not None:
        for passthrough in passthroughs:
            if type(passthrough) is np.ndarray:
                log.info("Passing np.ndarray at %s" % str(id(passthrough)))
                pt = passthrough
            else:
                log.info("Reading auxiliary file %s" % passthrough)
                pt = np.load(passthrough)
            names = [n for n in pt.dtype.names if n != 'uid' and n not in cs.dtype.names]
            if len(names) > 0:
                ptdf = util.dataframe_from_records_mapped(pt, {**general, **micrograph})
                ptdf = cryosparc_2_cs_particle_locations(pt, ptdf, swapxy=swapxy, invertx=invertx, inverty=inverty)
                # ptdf = cryosparc_2_cs_model_parameters(pt, ptdf, minphic=minphic)
                ptdf = cryosparc_2_cs_array_parameters(pt, ptdf)
                ptdf = cryosparc_2_cs_filament_parameters(pt, ptdf)
                key = star.UCSF.UID
                log.info("Trying to merge: %s" % ", ".join(names))
                fields = [c for c in ptdf.columns if c not in df.columns]
                log.info("Merging: %s" % ", ".join(fields))
                df = star.smart_merge(df, ptdf, fields=fields, key=key)
            else:
                log.info("This file contains no new information and will be ignored")

    if sys.version_info >= (3, 0):
        df = star.decode_byte_strings(df, fmt='UTF-8', inplace=True)

    # df[star.Relion.MAGNIFICATION] = 10000.0

    log.info("Directly copied fields: %s" % ", ".join(df.columns))

    if star.Relion.DEFOCUSANGLE in df:
        log.debug("Converting DEFOCUSANGLE from radians to degrees")
        df[star.Relion.DEFOCUSANGLE] = np.rad2deg(df[star.Relion.DEFOCUSANGLE])
    elif star.Relion.DEFOCUSV in df and star.Relion.DEFOCUSU in df:
        log.warning("Defocus angles not found")
    else:
        log.warning("Defocus values not found")

    if star.Relion.PHASESHIFT in df:
        log.debug("Converting PHASESHIFT from degrees to radians")
        df[star.Relion.PHASESHIFT] = np.rad2deg(df[star.Relion.PHASESHIFT])

    if star.Relion.ORIGINX in df.columns and boxsize is not None:
        df[star.Relion.ORIGINS] *= cs["blob/shape"][0] / boxsize

    if star.Relion.RANDOMSUBSET in df.columns:
        log.debug("Changing RANDOMSUBSET to 1-based index")
        if df[star.Relion.RANDOMSUBSET].value_counts().size == 1:
            df.drop(star.Relion.RANDOMSUBSET, axis=1, inplace=True)
        else:
            df[star.Relion.RANDOMSUBSET] += 1

    if star.Relion.CLASS in df.columns:
        log.debug("Changing CLASS to 1-based index")
        df[star.Relion.CLASS] += 1

    if star.Relion.OPTICSGROUP in df.columns:
        log.debug("Changing OPTICSGROUP to 1-based index")
        df[star.Relion.OPTICSGROUP] += 1

    if df.columns.intersection(star.Relion.ANGLES).size == len(star.Relion.ANGLES):
        log.debug("Converting Rodrigues coordinates to Euler angles")
        df[star.Relion.ANGLES] = np.rad2deg(geom.rot2euler(geom.expmap(df[star.Relion.ANGLES].values)))
        log.info("Converted Rodrigues coordinates to Euler angles")
    elif star.Relion.ANGLEPSI in df:
        log.debug("Converting ANGLEPSI from degrees to radians")
        df[star.Relion.ANGLEPSI] = np.rad2deg(df[star.Relion.ANGLEPSI])
    elif star.is_particle_star(df):
        log.warning("Angular alignment parameters not found")

    # for cases where the incoming .cs file was exported by CryoSPARC
    for path_field in [*star.Relion.PATH_FIELDS, *star.UCSF.PATH_FIELDS]:
        if path_field in df:
            if '>' in df[path_field].iloc[0]:
                df[path_field] = df[path_field].apply(lambda x: x.replace('>', '', 1))
                log.debug(f"Removed '>' from paths in field {path_field}")

    return df



import numpy as np
from PIL import Image
import torch
import os
import pickle
import random
from cryosparc.dataset import Dataset
from functools import partial
import multiprocessing
from tqdm import tqdm
from . import mrc
from . import fft



def sample_and_evaluate(path_list,save_path,num_stacks=5,num_particles=2000):
    if num_stacks> len(path_list):
        num_stacks=len(path_list)
    path_sampled=random.sample(path_list, num_stacks)
    np_image_raw_all=[]
    np_image_FT_all=[]
    imgs_len_all=[]

    for i in range(num_stacks):
        arr,_= mrc.parse_mrc(path_sampled[i])
        # np_image_raw = np.float32(mrcfile.open(path_sampled[i]).data)
        np_image_raw = np.float32(arr)
        imgs_len = np_image_raw.shape[0]


        particles = []
        for i in range(imgs_len):
            img = np_image_raw[i]
            particles.append(fft.ht2_center(img))

        np_image_FT = np.asarray(particles, dtype=np.float32)
        np_image_FT = fft.symmetrize_ht(np_image_FT)

        np_image_raw_all.append(np_image_raw)
        np_image_FT_all.append(np_image_FT)
        imgs_len_all.append(imgs_len)
    np_image_raw_all=np.concatenate(np_image_raw_all,axis=0)
    np_image_FT_all=np.concatenate(np_image_FT_all,axis=0)
    if num_particles> len(np_image_raw_all):
        num_particles=len(np_image_raw_all)
    index = np.random.choice(np.arange(len(np_image_raw_all)), size=num_particles, replace=False)
    np_image_raw_all_sampled=np_image_raw_all[index]
    np_image_FT_all_sampled=np_image_FT_all[index]
    means_raw=np.mean(np_image_raw_all_sampled)
    std_raw=np.std(np_image_raw_all_sampled)
    # means_FT=np.mean(np_image_FT_all_sampled)
    std_FT=np.std(np_image_FT_all_sampled)
    mean_imgs_len=np.mean(imgs_len_all)
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    with open(os.path.join(save_path, 'means_stds_raw.data'), 'wb') as filehandle:
        pickle.dump((means_raw, std_raw), filehandle)
    with open(os.path.join(save_path, 'means_stds_FT.data'), 'wb') as filehandle:
        pickle.dump((0.0, std_FT), filehandle)
    img_dim=np_image_raw_all.shape[-1]
    with open(os.path.join(save_path, 'img_dim.data'), 'wb') as filehandle:
        pickle.dump(img_dim, filehandle)
    # return (means_raw, std_raw), (means_FT, std_FT), mean_imgs_len
    return mean_imgs_len


def get_mean_std(path_list, Cnum=10000,is_normalize=True):
    # calculate means and stds
    imgs = []
    random.shuffle(path_list)
    if len(path_list) < Cnum:
        Cnum = len(path_list)
    sample_path_list = random.sample(path_list, Cnum)
    for i in range(Cnum):
        # img = np.array(Image.open(path_list[i]))
        img = pickle.load(open(sample_path_list[i], 'rb'))
        if isinstance(img,np.ndarray):
            min_val= np.min(img)
            max_val= np.max(img)
        else:
            min_val, max_val = img.getextrema()
            img = np.array(img)
        if max_val - min_val != 0 and is_normalize:
            img = (img - min_val) / (max_val - min_val)
        img = img.astype(np.float32)
        imgs.append(img)
    imgs_np = np.asarray(imgs)
    means = imgs_np.mean()
    stds = imgs_np.std()
    return means, stds


def sample_and_calculate_mean_std(path_list, Cnum=1000, Ctimes=1,is_normalize=True):
    mean_list = []
    std_list = []
    Cnum_all = 0
    for i in range(Ctimes):
        if Cnum_all >= len(path_list):
            break
        means, stds = get_mean_std(path_list, Cnum,is_normalize=is_normalize)
        mean_list.append(means)
        std_list.append(stds)
        Cnum_all += Cnum
    mean = np.mean(mean_list)
    std = np.mean(std_list)
    return mean, std

def mrcs_resize(mrcs, width, height, is_norm=False):
    resized_mrcs = np.zeros((mrcs.shape[0], width, height))
    # pbar = tqdm(range(mrcs.shape[0]))
    # pbar.set_description("resize mrcs to width*height")
    for i in range(mrcs.shape[0]):
        mrc = mrcs[i]
        # if is_norm:
        #     mrc = (mrc - np.min(mrc)) * (30 / (np.max(mrc) - np.min(mrc)))
        #     mrc = mrc - np.mean(mrc)
        mrc = Image.fromarray(mrc)
        resized_mrcs[i] = np.asarray(mrc.resize((width, height), Image.BICUBIC))
    resized_mrcs = resized_mrcs.astype('float32')
    return resized_mrcs

def mrcs_to_int8(mrcs):
    if torch.is_tensor(mrcs):
        new_mrcs = torch.zeros_like(mrcs, dtype=torch.uint8)
    else:
        new_mrcs = np.zeros_like(mrcs, dtype=np.uint8)
    for i in range(mrcs.shape[0]):
        new_mrcs[i] = to_int8(mrcs[i])
    return new_mrcs

# def to_int8(mrcdata):
#     # mrcdata = np.array(mrcdata)
#     # a=np.max(mrcdata)
#     # b=np.min(mrcdata)
#     if torch.is_tensor(mrcdata):
#         mrcdata = mrcdata.cpu().numpy()
#     if np.max(mrcdata) - np.min(mrcdata) != 0:
#         mrcdata_processed = (mrcdata - np.min(mrcdata)) / ((np.max(mrcdata) - np.min(mrcdata)))
#         mrcdata_processed = (mrcdata_processed * 255).astype(np.uint8)
#     else:
#         mrcdata_processed = mrcdata.astype(np.uint8)
#
#     return Image.fromarray(mrcdata_processed)
    # return mrcdata

def to_int8(mrcdata):
    if torch.is_tensor(mrcdata):
        mrcdata = (mrcdata - torch.min(mrcdata)) / (torch.max(mrcdata) - torch.min(mrcdata))
        mrcdata = (mrcdata * 255).type(torch.uint8)
    else:
        if np.max(mrcdata) - np.min(mrcdata) != 0:
            mrcdata = (mrcdata - np.min(mrcdata)) / ((np.max(mrcdata) - np.min(mrcdata)))
            mrcdata= (mrcdata * 255).astype(np.uint8)
        else:
            mrcdata = mrcdata.astype(np.uint8)
        mrcdata = Image.fromarray(mrcdata)
    return mrcdata


def window_mask(resolution, in_rad, out_rad=.99):
    assert resolution % 2 == 0
    x0, x1 = np.meshgrid(np.linspace(-1, 1, resolution, endpoint=False, dtype=np.float32),
                         np.linspace(-1, 1, resolution, endpoint=False, dtype=np.float32))
    r = (x0 ** 2 + x1 ** 2) ** .5
    mask = np.minimum(1.0, np.maximum(0.0, 1 - (r - in_rad) / (out_rad - in_rad)))
    return mask

def raw_csdata_process_from_cryosparc_dir(raw_data_path):
    if raw_data_path.endswith('cs'):
        raw_data_path = os.path.dirname(raw_data_path)
    new_csdata_path = os.path.join(raw_data_path, 'new_particles.cs')
    mrc_dir = raw_data_path

    # if not os.path.exists(new_csdata_path):
    passthrough_particles_path = None
    particles_cs_path = None
    other_cs_path = None
    for filename in os.listdir(raw_data_path):
        if filename.endswith('passthrough_particles.cs'):

            passthrough_particles_path = os.path.join(raw_data_path, filename)

        if filename.endswith('_split_0.cs'):

            passthrough_particles_path = os.path.join(raw_data_path, filename)

        if filename.endswith('extracted_particles.cs'):

            particles_cs_path = os.path.join(raw_data_path, filename)


        if filename.endswith('imported_particles.cs'):

            particles_cs_path = os.path.join(raw_data_path, filename)

        if filename.endswith('restacked_particles.cs'):

            particles_cs_path = os.path.join(raw_data_path, filename)

        if filename.endswith('split_0000.cs'):

            particles_cs_path = os.path.join(raw_data_path, filename)

        if filename.endswith('downsampled_particles.cs'):

            particles_cs_path = os.path.join(raw_data_path, filename)
        if filename.endswith('.cs'):
            other_cs_path = os.path.join(raw_data_path, filename)
    if not os.path.exists(new_csdata_path):
        if passthrough_particles_path is not None:
            cs_data = combine_cs_files_column(particles_cs_path, passthrough_particles_path)
        elif particles_cs_path is not None:
            cs_data = Dataset.load(particles_cs_path)

        elif other_cs_path is not None:
            cs_data = Dataset.load(other_cs_path)

        else:
            Exception(raw_data_path+': corresponding  not exists!')

        cs_data.save(new_csdata_path)
    else:
        cs_data = Dataset.load(new_csdata_path)

    if os.path.exists(os.path.join(raw_data_path, 'restack')):
        mrc_dir = os.path.join(raw_data_path, 'restack')
    elif os.path.exists(os.path.join(raw_data_path, 'extract')):
        mrc_dir = os.path.join(raw_data_path, 'extract')
    elif os.path.exists(os.path.join(raw_data_path, 'imported')):
        mrc_dir = os.path.join(raw_data_path, 'imported')
    elif os.path.exists(os.path.join(raw_data_path, 'downsample')):
        mrc_dir = os.path.join(raw_data_path, 'downsample')
    elif particles_cs_path is not None and particles_cs_path.endswith('split_0000.cs'):
        raw_dir='/'.join(raw_data_path.split('/')[0:-2])
        mrc_dir = raw_dir+'/'+'/'.join(cs_data['blob/path'][0].split('/')[0:-1])+'/'
    elif particles_cs_path is not None and particles_cs_path.endswith('downsampled_particles.cs'):
        raw_dir='/'.join(raw_data_path.split('/')[0:-1])
        mrc_dir=raw_dir
        # mrc_dir = raw_dir+'/'+'/'.join(cs_data['blob/path'][0].split('/')[0:-1])+'/'
    return cs_data,mrc_dir

def combine_cs_files_column(cs_path1, cs_path2):
    cs_data1=Dataset.load(cs_path1)
    cs_data2=Dataset.load(cs_path2)
    cs_data=Dataset.innerjoin(cs_data1,cs_data2)
    # save_dir='/'.join(save_path.split('/')[:-1])
    # if not os.path.exists(save_dir):
    #     os.makedirs(save_dir)
    # cs_data.save(save_path)
    # print('combined cs file saved in {}'.format(save_path))
    return cs_data
    # cs2star(save_path,save_path.replace('.cs','.star'))

def raw_data_preprocess_one_mrcs(name, mrc_dir, raw_dataset_save_dir, processed_dataset_save_dir, FT_dataset_save_dir,resize=224,
                                 is_to_int8=True,indeices_per_mrcs_dict=None):
    mrcs_path = os.path.join(mrc_dir, name)
    arr,_= mrc.parse_mrc(mrcs_path)
    # mrcs = np.float32(mrcfile.open(mrcs_path).data)
    mrcs = np.float32(arr)
    mrcs_len = mrcs.shape[0]
    raw_single_particle_path = []
    processed_single_particle_path = []
    FT_single_particle_path=[]
    if indeices_per_mrcs_dict is None:
        ids_list = range(mrcs_len)
    else:
        ids_list=indeices_per_mrcs_dict[name].tolist()

    if resize and mrcs.shape[1] != resize:

        processed_mrcs = mrcs_resize(mrcs, resize, resize)
    else:
        processed_mrcs = mrcs.copy()

    if is_to_int8:
        processed_mrcs = mrcs_to_int8(processed_mrcs)
        # processed_mrcs = to_int8(processed_mrcs)

    if FT_dataset_save_dir is not None:
        particles = []
        for i in range(mrcs_len):
            img= mrcs[i]
            particles.append(fft.ht2_center(img))

        FT_mrcs = np.asarray(particles, dtype=np.float32)
        FT_mrcs = fft.symmetrize_ht(FT_mrcs)

    for j in ids_list:

        n = str(j + 1).zfill(6)


        if not os.path.exists(os.path.join(processed_dataset_save_dir, name)):
            os.makedirs(os.path.join(processed_dataset_save_dir, name))
        with open(os.path.join(processed_dataset_save_dir, name, n + '.data'), 'wb') as filehandle:
            pickle.dump(Image.fromarray(processed_mrcs[j]).convert('L'), filehandle)

        processed_single_particle_path.append(os.path.join(processed_dataset_save_dir, name, n + '.data'))

        if raw_dataset_save_dir is not None:
            if not os.path.exists(os.path.join(raw_dataset_save_dir, name)):
                os.makedirs(os.path.join(raw_dataset_save_dir, name))
            with open(os.path.join(raw_dataset_save_dir, name, n + '.data'), 'wb') as filehandle:
                # pickle.dump(Image.fromarray(mrcs[j]), filehandle)
                pickle.dump(mrcs[j], filehandle)
            raw_single_particle_path.append(os.path.join(raw_dataset_save_dir, name, n + '.data'))
        else:
            raw_single_particle_path.append('')

        if FT_dataset_save_dir is not None:
            if not os.path.exists(os.path.join(FT_dataset_save_dir, name)):
                os.makedirs(os.path.join(FT_dataset_save_dir, name))
            with open(os.path.join(FT_dataset_save_dir, name, n + '.data'), 'wb') as filehandle:
                # pickle.dump(Image.fromarray(mrcs[j]), filehandle)
                pickle.dump(FT_mrcs[j], filehandle)
            FT_single_particle_path.append(os.path.join(FT_dataset_save_dir, name, n + '.data'))
        else:
            FT_single_particle_path.append('')

    return raw_single_particle_path, processed_single_particle_path,FT_single_particle_path

def raw_data_preprocess(raw_dataset_dir, dataset_save_dir, resize=224, is_to_int8=True,save_raw_data=True,save_FT_data=True,use_lmdb=True):

    if not os.path.exists(dataset_save_dir):
        os.makedirs(dataset_save_dir)

    cs_data, mrc_dir = raw_csdata_process_from_cryosparc_dir(raw_dataset_dir)


    if cs_data is not None:
        blob_path_list = cs_data['blob/path'].tolist()
        mrcs_names_list = [blob_path_list[i].split('/')[-1] for i in
                           range(len(blob_path_list))]
        # mrc_dir= os.path.dirname(mrc_list)
        mrcs_names_list_process = list(dict.fromkeys(mrcs_names_list))
        # mrcs_names_list=mrc_list

        print("Processing cs_data...")
        mrcs_names_np = np.array(mrcs_names_list)
        # Create a dictionary where the keys are names and the values are lists of indices
        blob_idx_np=_np = np.array(cs_data['blob/idx'].tolist())
        sorted_indices = np.argsort(mrcs_names_np)
        sorted_names = mrcs_names_np[sorted_indices]
        unique_names, counts = np.unique(sorted_names, return_counts=True)
        split_indices = np.split(sorted_indices, np.cumsum(counts)[:-1])
        indices_dict = dict(zip(unique_names, split_indices))
        indeices_per_mrcs_dict={}

        for name, indices in indices_dict.items():
            # Convert indices to numpy array
            indices_np = np.array(indices)

            # Get corresponding values in blob_idx_np
            values = blob_idx_np[indices_np]

            # Get the sorted indices based on the values
            sorted_indices = np.argsort(values)

            # Update indices in indices_dict in-place
            indices_dict[name] = indices_np[sorted_indices]
            indeices_per_mrcs_dict[name]=np.sort(values)


        func_append_data=partial(append_data,cs_data=cs_data,indices_dict=indices_dict)
        with multiprocessing.Pool(processes=8) as pool:
            results = pool.map(func_append_data, mrcs_names_list_process)
        new_cs_data = Dataset.append(results[0], *results[1:])




        new_csdata_path = os.path.join(dataset_save_dir, 'new_particles.cs')
        new_cs_data.save(new_csdata_path)
    else:
        indeices_per_mrcs_dict=None
        # mrcs_names_list_process=mrc_list
        new_cs_data=None

    if use_lmdb:
        tmp_data_lmdb_path = dataset_save_dir + '/lmdb_data'
        tmp_data_save_path= dataset_save_dir
        if not os.path.exists(tmp_data_lmdb_path):
            from src.data_preprocess.lmdb_preprocess import create_lmdb_dataset

            image_path_list = [os.path.join(mrc_dir, mrcs_name) for mrcs_name in mrcs_names_list_process]

            mean_len = sample_and_evaluate(image_path_list, tmp_data_save_path)

            # map_size = int(80 * 1024 * len(image_path_list) * 6)
            map_size = int(80 * 1024 * len(image_path_list) * mean_len * 4)
            # 创建 LMDB 数据库
            create_lmdb_dataset(image_path_list, tmp_data_lmdb_path, num_processes=8, chunksize=0,
                                map_size=map_size, window=False,  generate_ft_data=False,
                                save_raw_data=False)

    else:
        particles_dir_name = raw_dataset_dir.rstrip('/').split('/')[-1]

        if save_raw_data:
            raw_dataset_save_dir = os.path.join(dataset_save_dir, 'raw', particles_dir_name)
            if not os.path.exists(raw_dataset_save_dir):
                os.makedirs(raw_dataset_save_dir)
        else:
            raw_dataset_save_dir = None

        if save_FT_data:
            FT_dataset_save_dir = os.path.join(dataset_save_dir, 'FT', particles_dir_name)
            if not os.path.exists(FT_dataset_save_dir):
                os.makedirs(FT_dataset_save_dir)
        else:
            FT_dataset_save_dir = None

        processed_dataset_save_dir = os.path.join(dataset_save_dir, 'processed', particles_dir_name)
        if not os.path.exists(processed_dataset_save_dir):
            os.makedirs(processed_dataset_save_dir)
        raw_path_list = []
        processed_path_list = []
        FT_path_list = []


        phbar = tqdm(mrcs_names_list_process, desc='data preprocessing')
        func = partial(raw_data_preprocess_one_mrcs, mrc_dir=mrc_dir, raw_dataset_save_dir=raw_dataset_save_dir,FT_dataset_save_dir=FT_dataset_save_dir,
                       processed_dataset_save_dir=processed_dataset_save_dir, resize=resize, is_to_int8=is_to_int8,indeices_per_mrcs_dict=indeices_per_mrcs_dict)
        pool = multiprocessing.Pool(20)
        results = pool.map(func, phbar)
        pool.close()
        pool.join()

        for raw_single_particle_path, processed_single_particle_path,FT_single_particle_path in results:
            processed_path_list += processed_single_particle_path
            raw_path_list += raw_single_particle_path
            FT_path_list += FT_single_particle_path



        with open(os.path.join(dataset_save_dir, 'output_processed_tif_path.data'), 'wb') as filehandle:
            pickle.dump(processed_path_list, filehandle)

        with open(os.path.join(dataset_save_dir, 'output_tif_path.data'), 'wb') as filehandle:
            pickle.dump(raw_path_list, filehandle)

        mean_std_raw = sample_and_calculate_mean_std(raw_path_list,is_normalize=False)
        with open(dataset_save_dir + 'means_stds.data', 'wb') as filehandle:
            pickle.dump(mean_std_raw, filehandle)

        if FT_dataset_save_dir is not None:
            mean_std_FT = sample_and_calculate_mean_std(FT_path_list,is_normalize=False)
            with open(dataset_save_dir + 'means_stds_FT.data', 'wb') as filehandle:
                pickle.dump(mean_std_FT, filehandle)

    print('Cryoem data preprocess all done')
    return new_cs_data
def append_data(name,cs_data,indices_dict):
    # mm=np.sort(indices_dict[name])
    return cs_data.take(indices_dict[name])





# Copyright (C) 2017 Daniel Asarnow
# University of California, San Francisco
#
# Library functions for interesting plots.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import matplotlib.pyplot as plt
# import matplotlib as mpl
import numpy as np
import pandas as pd
import seaborn as sns

sns.set()
infinity = u'\u221e'
angstrom = u'\xc5'


def plot_fsc_curves(fsc, lgdtext=None, title=None, fname=None):
    # if fname is not None:
    #     mpl.rc("savefig", dpi=300)

    if type(fsc) is not list:
        fsc = [fsc]

    if type(fsc[0]) is str:
        fns = fsc
        fsc = []
        for fn in fns:
            fsc.append(pd.read_table(fn, header=None))
        for f in fsc:
            f.columns = ["freq", "fsc"]
            f["res"] = 1 / f["freq"]

    if lgdtext is None:
        lgdtext = ["Curve %d" % (i + 1) for i in range(len(fsc))]

    if title is None:
        title = "FSC Plot"

    sns.set(font_scale=3)
    fg, ax = plt.subplots(figsize=(20, 10))
    for f in fsc:
        f.plot(x="freq", y="fsc", ax=ax, legend=None, linewidth=4.)
    hl = ax.axhline(y=0.143, color='k', linestyle='-', linewidth=2.)
    hl.set_zorder(1)
    ax.set_xlim((np.min(fsc[0]["freq"]), np.max(fsc[0]["freq"])))
    ax.set_xlabel("Resolution (%s)" % angstrom)
    # ax.set_xticks([0., 1/20., 1/10., 1/6.7, 1/5., 1/4., 1/3.5, 1/3., 1/2.5])
    ax.set_xticks(np.arange(np.min(fsc[0]["freq"]), np.max(fsc[0]["freq"]), 0.05))
    # ax.set_xticklabels([infinity] + [("%.1f" % (1/f)).rstrip('0').rstrip('.') if (1/f) > 7 else "%.1f" % (1/f)
    #                                  for f in [1/20., 1/10., 1/6.7, 1/5., 1/4., 1/3.5, 1/3., 1/2.5]])
    ax.set_xticklabels([infinity] + [("%.1f" % f).rstrip('0').rstrip('.') if f > 7 else "%.1f" % f for f in
                                     1 / np.arange(0.05, np.max(fsc[0]["freq"]), 0.05)])
    ax.set_ylabel("Fourier Correlation")
    ax.set_ylim((np.min([f["fsc"] for f in fsc]), 1.02))
    ax.set_yticks(np.arange(0.0, 1.1, 0.2))
    ax.text(np.max(fsc[0]["freq"]), 0.155, "0.143", fontsize=36, horizontalalignment="right")
    ax.set_title(title)
    ax.legend(lgdtext, fontsize=24)
    if fname is not None:
        fg.savefig(fname, dpi=300)
        # mpl.rc("savefig", dpi=80)
    return fg, ax


def plot_angle_comparison(df1, df2, lgdtext=None, fname=None, maxrot=90):
    # if fname is not None:
    #     mpl.rc("savefig", dpi=300)

    if lgdtext is None:
        lgdtext = [u"Second (deg)", u"First (deg)"]

    sns.set(font_scale=3)
    f, ax = plt.subplots(1, 3, figsize=(30, 10))
    sns.regplot(df2["rlnAngleRot"], df1["rlnAngleRot"], fit_reg=False, scatter_kws={"s": 16}, ax=ax[0])
    ax[0].set_xlim((-maxrot, maxrot))
    ax[0].set_ylim((-maxrot, maxrot))
    ax[0].set_xticks(np.arange(-maxrot, maxrot+1, 15))
    ax[0].set_yticks(np.arange(-maxrot, maxrot+1, 15))
    ax[0].xaxis.label.set_visible(False)
    ax[0].set_ylabel(lgdtext[0])
    ax[0].set_title(u"$\phi$ ( $Z$ )", y=1.01)

    sns.regplot(df2["rlnAngleTilt"], df1["rlnAngleTilt"], fit_reg=False, scatter_kws={"s": 16}, ax=ax[1])
    ax[1].set_xlim((0, 180))
    ax[1].set_ylim((0, 180))
    ax[1].set_xticks(np.arange(0, 181, 30))
    ax[1].set_yticks(np.arange(0, 181, 30))
    ax[1].xaxis.label.set_visible(False)
    ax[1].yaxis.label.set_visible(False)
    ax[1].set_title(u"$\\theta$ ( $Y'$ )", y=1.01)

    sns.regplot(df2["rlnAnglePsi"], df1["rlnAnglePsi"], fit_reg=False, scatter_kws={"s": 16}, ax=ax[2])
    ax[2].set_xlim((-180, 180))
    ax[2].set_ylim((-180, 180))
    ax[2].set_xticks(np.arange(-180, 181, 45))
    ax[2].set_yticks(np.arange(-180, 181, 45))
    ax[2].xaxis.label.set_visible(False)
    ax[2].yaxis.label.set_visible(False)
    ax[2].set_title(u"$\psi$ ( $Z''$  )", y=1.01)
    f.text(0.5, -0.05, lgdtext[1], ha='center', fontsize=36)
    f.tight_layout(pad=1., w_pad=-1.5, h_pad=0.5)
    if fname is not None:
        f.savefig(fname, dpi=300)
        # mpl.rc("savefig", dpi=80)
    return f, ax



# Copyright (C) 2017-2018 Daniel Asarnow
# University of California, San Francisco
#
# Library functions for volume data.
# See README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from scipy.ndimage import binary_dilation
from scipy.ndimage import binary_fill_holes
from scipy.ndimage import distance_transform_edt
from scipy.ndimage import label
from scipy.ndimage import labeled_comprehension


def binary_sphere(r, le=True):
    rr = np.linspace(-r, r, 2 * r + 1)
    x, y, z = np.meshgrid(rr, rr, rr)
    if le:
        sph = (x ** 2 + y ** 2 + z ** 2) <= r ** 2
    else:
        sph = (x ** 2 + y ** 2 + z ** 2) < r ** 2
    return sph


def binary_volume_opening(vol, minvol):
    if minvol == 0:
        return vol.copy()
    lb_vol, num_objs = label(vol)
    lbs = np.arange(1, num_objs + 1)
    v = labeled_comprehension(lb_vol > 0, lb_vol, lbs, np.sum, int, 0)
    if minvol < 0:
        ix = np.isin(lb_vol, lbs[np.argsort(v)[minvol:]])
    else:
        ix = np.isin(lb_vol, lbs[v >= minvol])
    newvol = np.zeros(vol.shape, dtype=np.bool)
    newvol[ix] = vol[ix]
    return newvol


def binary_dilate(vol, size, strel=False, dt=None):
    if size == 0:
        return vol.copy()
    if not strel:
        if dt is None:
            dt = distance_transform_edt(~vol)
        return vol | (dt <= size)
    else:
        se = binary_sphere(size, False)
        return binary_dilation(vol, structure=se, iterations=1)


def binarize_volume(vol, t, minvol=0, fill=False):
    mask = vol >= t
    if minvol != 0:
        mask = binary_volume_opening(mask, minvol)
    if fill:
        mask = binary_fill_holes(mask)
    return mask




# @Author : Written/modified by Yang Yan
# @Time   : 2022/6/20 下午7:08
# @E-mail : yanyang98@yeah.net
# @Function :
import os
import numpy as np
from torchvision.utils import save_image
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from scipy import stats
from sklearn.metrics import mean_squared_error, r2_score


# def calculate_interval_proportions(input_dict, num_intervals=10):
#     # 获取所有的值
#     values = list(input_dict.values())
#
#     # 计算最小值和最大值
#     min_value = min(values)
#     max_value = max(values)
#
#     # 计算区间宽度
#     interval_width = (max_value - min_value) / num_intervals
#
#     # 初始化一个字典来存储每个区间的计数
#     interval_counts = {i: 0 for i in range(num_intervals)}
#
#     # 遍历所有值，计算它们所在的区间
#     for value in values:
#         if value == max_value:
#             interval_counts[num_intervals - 1] += 1
#         else:
#             interval_index = int((value - min_value) // interval_width)
#             interval_counts[interval_index] += 1
#
#     # 计算每个区间的比例
#     total_values = len(values)
#     interval_proportions = {i: count / total_values for i, count in interval_counts.items()}
#
#     # 创建一个新字典，key与输入一致，value为对应值所在区间的比例
#     result_dict = {}
#     for key, value in input_dict.items():
#         if value == max_value:
#             interval_index = num_intervals - 1
#         else:
#             interval_index = int((value - min_value) // interval_width)
#         result_dict[key] = interval_proportions[interval_index]
#
#     return result_dict

def calculate_interval_proportions(input_dict, num_intervals=10):
    # 获取所有的值
    values = list(input_dict.values())
    values_sum=sum(values)
    values_num=len(values)


    # 创建一个新字典，key与输入一致，value为对应值所在区间的比例
    result_dict = {}
    for key, value in input_dict.items():
        result_dict[key] = values_num*value/values_sum

    return result_dict

def get_mean_error_distribution(mean_error_dict):
    mean_error_good_dict={name:mean_error for name,mean_error in mean_error_dict.items() if not name.lower().endswith('_bad')}
    mean_error_bad_dict={name:mean_error for name,mean_error in mean_error_dict.items() if name.lower().endswith('_bad')}
    dis_good=calculate_interval_proportions(mean_error_good_dict)
    dis_bad=calculate_interval_proportions(mean_error_bad_dict)
    # mean_error_dict_all={'good':dis_good,'bad':dis_bad}
    mean_error_dict_all={**dis_good,**dis_bad}
    return mean_error_dict_all

def balanced_mask(target_y_t_score, bar_score,random_seed=0):
    np.random.seed(random_seed)
    # 计算满足两个条件的掩码
    high_mask = target_y_t_score > bar_score[1]
    low_mask = target_y_t_score <= bar_score[0]

    # 计算满足每个条件的元素数量
    high_count = np.sum(high_mask)
    low_count = np.sum(low_mask)

    # 确定需要保留的元素数量
    keep_count = min(high_count, low_count)
    if keep_count==0:
        return np.ones_like(target_y_t_score, dtype=bool)

    # 如果高分数的元素过多，随机选择部分元素
    if high_count > keep_count:
        high_indices = np.where(high_mask)[0]
        keep_high_indices = np.random.choice(high_indices, keep_count, replace=False)
        high_mask = np.zeros_like(high_mask)
        high_mask[keep_high_indices] = True

    # 如果低分数的元素过多，随机选择部分元素
    if low_count > keep_count:
        low_indices = np.where(low_mask)[0]
        keep_low_indices = np.random.choice(low_indices, keep_count, replace=False)
        low_mask = np.zeros_like(low_mask)
        low_mask[keep_low_indices] = True

    # 合并两个条件的结果
    mask = high_mask | low_mask

    return mask
def evaluate_fit(score_predicted_data, score_truth_data):
    """
    评估预测数据与真实数据对 y=x 的拟合程度。

    参数:
    confidence_predicted_data : np.ndarray
        预测的置信度数据
    confidence_truth_data : np.ndarray
        真实的置信度数据

    返回:
    dict
        包含多个评估指标的字典
    """
    # 确保输入数据的形状相同
    assert score_predicted_data.shape == score_truth_data.shape, "输入数组的形状必须相同"

    # 计算均方误差 (MSE)
    mse = mean_squared_error(score_truth_data, score_predicted_data)

    # 计算均方根误差 (RMSE)
    rmse = np.sqrt(mse)

    # 计算决定系数 (R-squared)
    r2 = r2_score(score_truth_data, score_predicted_data)

    # 计算皮尔逊相关系数
    pearson_corr, _ = stats.pearsonr(score_truth_data.flatten(), score_predicted_data.flatten())

    # 计算斯皮尔曼等级相关系数
    spearman_corr, _ = stats.spearmanr(score_truth_data, score_predicted_data)

    # 计算平均绝对误差 (MAE)
    mae = np.mean(np.abs(score_predicted_data - score_truth_data))

    # 计算最大绝对误差
    max_error = np.max(np.abs(score_predicted_data - score_truth_data))

    return {
        "MSE": mse,
        "RMSE": rmse,
        "R-squared": r2,
        "Pearson Correlation": pearson_corr,
        "Spearman Correlation": spearman_corr,
        "MAE": mae,
        "Max Error": max_error
    }
def plot_mean_score_histogram(score_predicted_data, score_truth_data, result_dir, epoch='final'):
    # 确保输入数据的形状正确
    assert score_predicted_data.shape == score_truth_data.shape
    assert len(score_predicted_data.shape) == 1 or score_predicted_data.shape[1] == 1

    # 如果是2D数组，转换为1D
    if len(score_predicted_data.shape) == 2:
        score_predicted_data = score_predicted_data.flatten()
        score_truth_data = score_truth_data.flatten()

    # 创建10个区间
    bins = np.linspace(0, 1, 11)

    # 计算每个区间的平均真实置信度和标准差
    averages = []
    std_devs = []
    for i in range(10):
        mask = (score_predicted_data >= bins[i]) & (score_predicted_data < bins[i + 1])
        if np.sum(mask) > 0:
            avg = np.mean(score_truth_data[mask])
            std = np.std(score_truth_data[mask])  # 计算标准差
        else:
            avg = 0
            std = 0
        averages.append(avg)
        std_devs.append(std)

    # 创建图表对象
    fig, ax = plt.subplots(figsize=(10, 6))

    # 绘制柱状图和误差线
    # bars = ax.bar(bins[:-1], averages, width=0.08, align='edge', yerr=std_devs, capsize=5, ecolor='red', alpha=0.7)
    bars = ax.bar(bins[:-1], averages, width=0.08, align='edge',  capsize=5, ecolor='red', alpha=0.7)

    # 设置标签和标题
    ax.set_xlabel('Predicted Score')
    ax.set_ylabel('Average True Score')
    # ax.set_title('Average True Confidence vs Predicted Confidence (Error bars: Standard Deviation)')
    ax.set_title('Average True Score vs Predicted Score')

    # 设置坐标轴范围和刻度
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_xticks(np.arange(0, 1.1, 0.1))

    # 添加网格线
    ax.grid(axis='y', linestyle='--', alpha=0.7)

    # 在每个柱子上方显示具体数值
    for bar in bars:
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width() / 2., height,
                f'{height:.2f}',
                ha='center', va='bottom')

    plt.savefig(result_dir + '/figures/mean_score_interval_' + str(epoch) + '.svg')
    plt.close()
    return fig
def plot_mean_score_violin(scores_data_predicted_np, scores_data_true_np, result_dir, epoch='final'):
    fig, ax = plt.subplots(figsize=(8, 6))

    # 设置字体大小
    plt.rcParams.update({'font.size': 14})

    # 设置x轴刻度
    x_ticks = np.arange(0, 1.1, 0.1)
    # 存储每个区间的平均值
    mean_values = []
    # 存储小提琴图的位置
    violin_positions = []

    # 遍历每个x轴刻度范围
    for i, x_tick in enumerate(x_ticks[:-1]):
        next_x_tick = x_ticks[i + 1]
        mid_point = (x_tick + next_x_tick) / 2
        violin_positions.append(mid_point)

        if i == len(x_ticks) - 2:
            mask = (scores_data_true_np >= x_tick) & (scores_data_true_np <= next_x_tick)
        else:
            mask = (scores_data_true_np >= x_tick) & (scores_data_true_np < next_x_tick)

        # 获取当前范围内的预测值
        predicted_values = scores_data_predicted_np[mask]

        # 计算当前范围内的平均值
        mean_value = np.mean(predicted_values)
        mean_values.append(mean_value)

        # 绘制小提琴图
        if len(predicted_values) > 0:
            violin_parts = ax.violinplot([predicted_values], positions=[mid_point], showmeans=False, showmedians=False,
                                         showextrema=False, widths=0.04)



            # 设置小提琴图颜色和透明度
            for vp in violin_parts['bodies']:
                vp.set_facecolor('lightblue')
                vp.set_edgecolor('none')
                vp.set_alpha(0.7)

    # 绘制折线图
    ax.plot(violin_positions, mean_values, color='red', marker='o')

    # 在节点上显示平均值数字
    for x, y in zip(violin_positions, mean_values):
        ax.annotate(f'{y:.2f}', (x, y), xytext=(5, 5), textcoords='offset points')

    # 设置x轴刻度标签
    ax.set_xticks(x_ticks)
    ax.set_xticklabels([f'{x_tick:.1f}' for x_tick in x_ticks])

    # 设置y轴范围
    ax.set_ylim(0, 1.1)

    # 设置坐标轴标签和标题
    ax.set_xlabel('True Score')
    ax.set_ylabel('Predicted Score')
    ax.set_title('Violin Plot')

    plt.savefig(result_dir + '/figures/mean_score_interval_violin_' + str(epoch) + '.svg')
    plt.close()
    return fig
def plot_score_distribution(score_data, result_dir, epoch='final', score_bar=[0.4, 0.7]):
    # 确保输入数据是一维数组
    score_data = score_data.flatten()

    # 创建直方图数据
    hist, bin_edges = np.histogram(score_data, bins=np.arange(0, 1.1, 0.1))

    # 计算每个区间的数量
    low = np.sum(score_data < score_bar[0])
    mid = np.sum((score_data >= score_bar[1]) & (score_data < score_bar[0]))
    high = np.sum(score_data >= score_bar[1])

    # 创建自定义颜色映射
    colors = ['#3498db', '#f39c12', '#2ecc71']  # 蓝色、橙色、绿色
    n_bins = 3
    cmap = LinearSegmentedColormap.from_list('custom', colors, N=n_bins)

    # 为每个柱子分配颜色
    bar_colors = [cmap(0)] * 10 + [cmap(1)] * 1 + [cmap(2)] * 1
    for i in range(10):
        if bin_edges[i] >= score_bar[1] and bin_edges[i] < score_bar[0]:
            bar_colors[i] = cmap(1)
        if bin_edges[i] >= score_bar[1]:
            bar_colors[i] = cmap(2)

    # 创建图形和轴
    fig, ax = plt.subplots(figsize=(12, 6))

    # 绘制直方图
    bars = ax.bar(bin_edges[:-1], hist, width=0.08, align='edge', color=bar_colors, edgecolor='white')

    # 添加标签和标题
    ax.set_xlabel('Score', fontsize=12)
    ax.set_ylabel('Count', fontsize=12)
    ax.set_title(f'Score Distribution (Bar: {score_bar})', fontsize=14)

    # 在每个柱子上方显示数量
    for bar in bars:
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width() / 2., height,
                f'{int(height)}',
                ha='center', va='bottom', fontsize=8)

    # 添加图例
    if score_bar[0] != score_bar[1]:
        legend_elements = [
            plt.Rectangle((0, 0), 1, 1, facecolor=cmap(0), edgecolor='none', label=f'Low: {low}'),
            plt.Rectangle((0, 0), 1, 1, facecolor=cmap(1), edgecolor='none', label=f'Mid: {mid}'),
            plt.Rectangle((0, 0), 1, 1, facecolor=cmap(2), edgecolor='none', label=f'High: {high}')
        ]
    else:
        legend_elements = [
            plt.Rectangle((0, 0), 1, 1, facecolor=cmap(0), edgecolor='none', label=f'Low: {low}'),
            plt.Rectangle((0, 0), 1, 1, facecolor=cmap(2), edgecolor='none', label=f'High: {high}')
        ]
    ax.legend(handles=legend_elements, loc='upper left', bbox_to_anchor=(0.01, 1), fontsize=10)

    # 调整x轴刻度
    ax.set_xticks(np.arange(0, 1.1, 0.1))
    ax.set_xlim(-0.23, 1.05)  # 稍微扩大x轴范围，以便柱子不会紧贴边缘

    # 显示网格线
    ax.grid(axis='y', linestyle='--', alpha=0.7)

    # 调整图形布局
    plt.tight_layout()

    # 调整左边距和上边距
    plt.subplots_adjust(left=0.08, top=0.9)
    plt.savefig(result_dir + '/figures/num_interval_' + str(epoch) + '.svg')
    plt.close()
    return fig
def plot_score_interval_precision_recall(score_predicted_data, score_truth_data, result_dir, epoch='final', score_bar=[0.4, 0.7]):
    """
    绘制confidence_predicted_data在各个confidence间隔内二分类预测的准确度柱状图

    参数：
    - confidence_predicted_data (ndarray): n*1大小的预测置信度数据
    - confidence_truth_data (ndarray): n*1大小的真实标签数据
    - result_dir (str): 结果保存目录
    - epoch (str): 当前epoch，默认为'final'
    - confidence_bar (float): 0.5到1之间的数字，用于区分置信度的阈值
    """
    # 将confidence_predicted_data和confidence_truth_data转换为一维数组
    score_predicted_data = score_predicted_data.flatten()
    score_truth_data = score_truth_data.flatten()

    # 计算每个置信度间隔内的预测准确度和召回率
    bins = np.arange(0, 1.1, 0.1)
    precision = []
    recall = []
    counts = []
    for i in range(len(bins) - 1):
        lower = bins[i]
        upper = bins[i + 1]
        if i == len(bins) - 2:
            upper += 0.01
        mask = (score_predicted_data >= lower) & (score_predicted_data < upper)

        if mask.sum() > 0:
            predicted_labels = (score_predicted_data[mask] >= score_bar[1]).astype(int)
            truth_labels = (score_truth_data[mask] >= score_bar[1]).astype(int)
            tp = np.sum(predicted_labels == truth_labels)


            precision.append(tp / predicted_labels.shape[0] if predicted_labels.shape[0] > 0 else 0)

            counts.append(mask.sum())
        else:
            precision.append(0)
            counts.append(0)
        mask_truth = (score_truth_data >= lower) & (score_truth_data < upper)
        if mask_truth.sum() > 0:
            predicted_labels2 = (score_predicted_data[mask_truth] >= score_bar[1]).astype(int)
            truth_labels2 = (score_truth_data[mask_truth] >= score_bar[1]).astype(int)
            tp2 = np.sum(predicted_labels2 == truth_labels2)
            truth_labels_interval=(score_truth_data[mask_truth] >= score_bar[1]).astype(int)
            recall.append(tp2 / truth_labels_interval.shape[0] if truth_labels_interval.shape[0] > 0 else 0)
        else:
            recall.append(0)

    # 创建自定义颜色映射
    colors = ['#3498db', '#f39c12', '#2ecc71']  # 蓝色、橙色、绿色
    n_bins = 2
    cmap = LinearSegmentedColormap.from_list('custom', colors, N=n_bins)

    # 创建一个fig对象
    fig, ax = plt.subplots(figsize=(12, 6))

    # 绘制柱状图
    bar_width = 0.04
    x = (bins[:-1] + bins[1:]) / 2  # 计算每个区间的中点
    rects1 = ax.bar(x - bar_width/2, precision, width=bar_width, alpha=0.7, color=cmap(0), label='Precision')
    rects2 = ax.bar(x + bar_width/2, recall, width=bar_width, alpha=0.7, color=cmap(1), label='Recall')

    def autolabel(rects):
        for rect in rects:
            height = rect.get_height()
            ax.annotate(f'{height:.2f}',
                        xy=(rect.get_x() + rect.get_width() / 2, height),
                        xytext=(0, 3),  # 3 points vertical offset
                        textcoords="offset points",
                        ha='center', va='bottom', fontsize=8)

    autolabel(rects1)
    autolabel(rects2)

    # 为大于等于confidence_bar的柱子添加黑色边框
    for i in range(len(bins) - 1):
        if bins[i] >= score_bar[1]:
            ax.bar(x[i] - bar_width / 2, precision[i], width=bar_width, alpha=0.7, color='none', edgecolor='black',
                   linewidth=2)
            ax.bar(x[i] + bar_width / 2, recall[i], width=bar_width, alpha=0.7, color='none', edgecolor='black',
                   linewidth=2)

    # 添加标签和标题
    ax.set_xlabel('Score')
    ax.set_ylabel('Precision / Recall')
    ax.set_title('Precision and Recall vs Score')
    ax.set_ylim(0, 1.1)  # 设置y轴范围避免重合
    ax.set_xticks(bins)  # 设置x轴刻度以0.1为间隔
    ax.set_xticklabels([f'{b:.1f}' for b in bins])  # 设置x轴刻度标签

    # 添加网格线
    ax.grid(True, linestyle='--', alpha=0.7)

    # 添加图例
    ax.legend()

    # 显示图形
    plt.tight_layout()
    plt.savefig(result_dir + f'/figures/score_interval_precision_recall_{epoch}.svg', dpi=300, bbox_inches='tight')
    plt.close(fig)  # 关闭图形，释放内存
    return fig
def plot_score_interval_inf(score_data, labels_p, labels_t, result_dir, epoch='final', bin_num=10):
    labels_data = []
    accuracy_data = []
    precision_data = []
    interval_num_data = []
    interval_pos_num_data = []

    if labels_t is None:
        calculate_acc = False
    elif -1 in labels_t:
        calculate_acc = False
    else:
        calculate_acc = True
    for i in range(bin_num):
        score_min = i / bin_num
        score_max = (i + 1) / bin_num
        my_index = (score_data >= score_min) & (score_data < score_max)
        # confidence_data_i = confidence_data[my_index]
        labels_p_i = labels_p[my_index]

        interval_num_data.append(len(labels_p_i))
        pos_num = sum(labels_p_i == 1)
        interval_pos_num_data.append(pos_num)
        labels_data.append(str(score_min) + '-' + str(score_max))

        if calculate_acc and len(labels_p_i) > 0:
            labels_t_i = labels_t[my_index]
            accuracy_data_i = np.mean(labels_p_i == labels_t_i)
            accuracy_data.append(accuracy_data_i)
            precision_correct_num = sum(labels_p_i[labels_p_i == 1] == labels_t_i[labels_p_i == 1])
            if pos_num>0:
                precision_data.append((precision_correct_num / pos_num))
            else:
                precision_data.append(0)
        else:
            accuracy_data.append(0)
            precision_data.append(0)
    if not os.path.exists(result_dir + '/figures/'):
        os.makedirs(result_dir + '/figures/')
    my_width = 0.4
    x = np.arange(bin_num)
    fig_num, ax_num = plt.subplots()
    fig_num.set_size_inches((14, 8))

    num_bar = ax_num.bar(x - my_width / 2, interval_num_data, my_width,
                         label='particles number (all ' + str(len(labels_p)) + ')')
    ax_num.bar_label(num_bar, label_type='edge')

    num_bar_pos = ax_num.bar(x + my_width / 2, interval_pos_num_data, my_width,
                             label='positive particles number (all ' + str(sum(interval_pos_num_data)) + ')')
    ax_num.bar_label(num_bar_pos, label_type='edge')
    ax_num.set_xticks(x)
    ax_num.set_xticklabels(labels_data)
    ax_num.set_xlabel('score')
    ax_num.set_title('Particles num statistics')
    ax_num.legend()
    fig_num.savefig(result_dir + '/figures/score_interval_num_' + str(epoch) + '.svg')

    if calculate_acc:
        fig_acc, ax_acc = plt.subplots()
        fig_acc.set_size_inches((14, 8))
        acc_bar = ax_acc.bar(x - my_width / 2, accuracy_data, my_width, label='classification accuracy')
        ax_acc.bar_label(acc_bar, labels=['%.3f' % acc if acc != 0 else str(0) for acc in accuracy_data],
                         label_type='edge')
        p_bar = ax_acc.bar(x + my_width / 2, precision_data, my_width, label='classification precision')
        ax_acc.bar_label(p_bar, labels=['%.3f' % p if p != 0 else str(0) for p in precision_data], label_type='edge')
        ax_acc.set_xticks(x)
        ax_acc.set_xticklabels(labels_data)
        ax_acc.set_xlabel('score')
        ax_acc.set_title('Classification performance')
        ax_acc.legend()
        fig_acc.savefig(result_dir + '/figures/score_interval_acc_' + str(epoch) + '.svg')
    else:
        fig_acc = None
    plt.close(fig_num)
    if fig_acc is not None:
        plt.close(fig_acc)

    return fig_num, fig_acc


class AverageMeter(object):
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def save_clustering_labels(labels_path, epoch, clustering_labels):
    if not os.path.exists(labels_path):
        os.makedirs(labels_path + '/')
    np.save(labels_path + '/predict_labels_epoch' + str(epoch) + '.npy', clustering_labels)


def save_acc_data(acc, acc_best, nmi, nmi_best, tb_writer, acc_sum, nmi_sum, clustering_times, epoch, out_path):
    if acc > acc_best:
        acc_best = acc
    if nmi > nmi_best:
        nmi_best = nmi
    acc_sum = acc_sum + acc
    nmi_sum = nmi + nmi_sum
    clustering_times = clustering_times + 1
    tb_writer.add_scalar("accuracy:", acc, epoch)
    tb_writer.add_scalar("nmi:", nmi, epoch)
    tb_writer.add_scalar("mean accuracy:", acc_sum / clustering_times,
                         clustering_times)
    tb_writer.add_scalar("mean NMI:", nmi_sum / clustering_times,
                         clustering_times)
    if not os.path.exists(out_path + 'acc_data/'):
        os.makedirs(out_path + 'acc_data/')
    with open(out_path + 'acc_data/' + 'acc_data.txt', 'w') as average_acc:
        average_acc.write('best accuracy' + str(acc_best))
        average_acc.write('\nbest NMI' + str(nmi_best))
        average_acc.write('\naverage accuracy' + str(acc_sum / clustering_times))
        average_acc.write('\naverage NMI' + str(nmi_sum / clustering_times))
    return acc_best, nmi_best, acc_sum, nmi_sum, clustering_times


def save_trained_model(model, optimizer, epoch, save_path):
    import torch
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    torch.save({'optimizer': optimizer.state_dict(), 'model': model.state_dict(),
                'epoch': epoch + 1}, save_path + 'epoch' + str(epoch) + '_model.pth.tar')


def save_averages(average_imgs_all, average_generated_imgs=None, run_root_dir=None, epoch=0):
    import torch
    import mrcfile
    if not os.path.exists(run_root_dir + 'averages/'):
        os.makedirs(run_root_dir + 'averages/')
    save_image(torch.unsqueeze(torch.from_numpy(average_imgs_all), 1),
               run_root_dir + 'averages/clustering_result_' + str(epoch) + '.png')
    if average_generated_imgs is not None:
        save_image(torch.unsqueeze(torch.from_numpy(average_generated_imgs), 1),
                   run_root_dir + 'averages/generated_clustering_result_' + str(epoch) + '.png')
        # projectons_file = mrcfile.new(
        #     run_root_dir + 'averages/generated_clustering_averages_' + str(epoch) + '.mrcs',
        #     average_generated_imgs, overwrite=True)
    projectons_file = mrcfile.new(
        run_root_dir + 'averages/clustering_averages_' + str(epoch) + '.mrcs',
        average_imgs_all, overwrite=True)
    projectons_file.close()


def confusion_matrix(predictions, gt, class_names, output_file=None):
    # Plot confusion_matrix and store result to output_file
    import sklearn.metrics
    import matplotlib.pyplot as plt
    confusion_matrix = sklearn.metrics.confusion_matrix(gt, predictions)
    confusion_matrix = confusion_matrix / np.sum(confusion_matrix, 1)

    fig, axes = plt.subplots(1)
    plt.imshow(confusion_matrix, cmap='Blues')
    axes.set_xticks([i for i in range(len(class_names))])
    axes.set_yticks([i for i in range(len(class_names))])
    axes.set_xticklabels(class_names, ha='right', fontsize=8, rotation=40)
    axes.set_yticklabels(class_names, ha='right', fontsize=8)

    for (i, j), z in np.ndenumerate(confusion_matrix):
        if i == j:
            axes.text(j, i, '%d' % (100 * z), ha='center', va='center', color='white', fontsize=6)
        else:
            pass

    plt.tight_layout()
    if output_file is None:
        plt.show()
    else:
        plt.savefig(output_file, dpi=300, bbox_inches='tight')
    plt.close()


def free_mem():
    import os
    result = os.popen("fuser -v /dev/nvidia*").read()
    results = result.split()
    for pid in results:
        os.system(f"kill -9 {int(pid)}")


def numsCheng(i):
    for m in range(100000):
        m = i * 2
        pass
    return i * 2, 2


def multi_process_test():
    import torch
    import time
    from accelerate import Accelerator
    from accelerate.utils import InitProcessGroupKwargs
    from multiprocessing.pool import Pool

    print("start")
    time1 = time.time()
    nums_list = range(100000)
    pool = Pool(processes=10)
    result = pool.map(numsCheng, nums_list)
    pool.close()  # 关闭进程池，不再接受新的进程
    pool.join()  # 主进程阻塞等待子进程的退出
    print("end")
    # print(result)
    time2 = time.time()
    print("计算用时：", time2 - time1)

    # for i,j in result:
    #     print(i,j)


if __name__ == '__main__':
    a=[0.01,0.12,0.1,0.13,0.2,0.3,0.4,0.5,0.6,0.7,0.8]
    b=[0.0001,0.01,0.005,0.12,0.1,0.13,0.2,0.3,0.4,0.5,0.6,0.7,0.8]
    bar=0.6
    mask=balanced_mask(np.array(a),bar)
    mask2=balanced_mask(np.array(b),bar)
    mask3=balanced_mask(np.array(a),bar,random_seed=2)
    pass
    # multi_process_test()
    # import matplotlib.pyplot as plt
    # import numpy as np
    # import seaborn as sns
    # import pandas as pd
    #
    # bin_num = 10
    # confidence_data = np.random.rand(100)
    # labels_p = np.random.randint(0, 10, 100)
    # labels_t = np.random.randint(0, 10, 100)
    # # labels_p = np.random.randint(0, 10, 100)
    # # labels_t = np.random.randint(0, 10, 100)
    # # confidence_data = confidence_data[:, labels_p == labels_t]
    # labels_p = labels_p[labels_p == labels_t]
    # # confidence_data = np.max(confidence_data, axis=1)
    # confidence_data = np.clip(confidence_data, 0, 1)
    # bins = np.linspace(0, 1, bin_num)
    # labels = np.digitize(confidence_data, bins)
    # data = pd.DataFrame({'confidence': confidence_data, 'label': labels_p, 'bin': labels})
    # data = data.groupby('bin').apply(lambda x: np.mean(x['label'] == x['label'].mode().values[0]))
    # data = data.reset_index()
    # data.columns = ['bin', 'accuracy']
    # plt.figure()
    # sns.barplot(x='bin', y='accuracy', data=data)
    # # plt.show()
    # plt.savefig('/yanyang2/projects/results/particle_classification/inference_test/2024_2_25_P954W7J29/test.png')

    # free_mem()
    # multi_process_test()
    # l=['a','b','c']
    # l.append(None)
    # print(l)
    # print('a','b')
# def save_classified_particles(run_root_dir, epoch, generated_imgs, clustering_labels, class_number_arry,cluster_num):
#     if not os.path.exists(run_root_dir + 'class_single_particles/epoch' + str(epoch) + '/'):
#         os.makedirs(run_root_dir + 'class_single_particles/epoch' + str(epoch) + '/')
#     if not os.path.exists(run_root_dir + 'generated_class_single_particles/epoch' + str(epoch) + '/'):
#         os.makedirs(run_root_dir + 'generated_class_single_particles/epoch' + str(epoch) + '/')
#     if Running_Paras.is_save_clustered_single_particles:
#         for i in range(cluster_num):
#             # class_single_particles = raw_mrcArrays[clustering_labels == i]
#             generated_class_single_particles = generated_imgs[clustering_labels == i]
#             generated_class_single_particles = np.squeeze(generated_class_single_particles)
#             class_particles = mrcfile.new(
#                 run_root_dir + 'generated_class_single_particles/epoch' + str(epoch) + '/' + 'class_' + str(
#                     i) + "_" + str(class_number_arry[i]) + '.mrcs',
#                 generated_class_single_particles, overwrite=True)
#             class_particles.close()
#         evaluate_utils.classify_mrcs(clustering_labels, cluster_num,
#                                      Running_Paras.path_for_data_classify,
#                                      run_root_dir + 'class_single_particles/epoch' + str(epoch) + '/')



# Copyright (C) 2018 Daniel Asarnow
# University of California, San Francisco
#
# Library of miscellaneous utility functions.
# See help text and README file for more information.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from .quat_numba import distq
from .quat_numba import qslerp
from .quat_numba import qtimes
from .quat import meanq


def double_center(arr, reference=None, inplace=False):
    assert arr.ndim == 2
    if reference is None:
        mu0 = np.mean(arr, axis=0, keepdims=True)
        mu = np.mean(arr)
    else:
        mu0 = np.mean(reference, axis=0, keepdims=True)
        mu = np.mean(reference)
    mu1 = np.mean(arr, axis=1, keepdims=True)
    if inplace:
        arr -= mu0
        arr -= mu1
        arr += mu
    else:
        arr = arr - mu0
        arr -= mu1
        arr += mu
    return arr


def isrotation(r, tol=1e-6):
    """Check for valid rotation matrix"""
    check = np.identity(3, dtype=r.dtype) - np.dot(r.T, r)
    if tol is None:
        return check
    return np.linalg.norm(check) < tol


def qslerp_mult_balanced(keyq, steps_per_deg=10):
    qexp = []
    for i in range(0, len(keyq) - 1):
        ninterp = np.round(np.rad2deg(2 * distq(keyq[i], keyq[i + 1])) * steps_per_deg)
        for t in np.arange(0, 1, 1 / ninterp):
            qexp.append(qslerp(keyq[i], keyq[i + 1], t))
    qexp.append(keyq[-1])
    return np.array(qexp)


def findkeyq(qarr, kpcs, nkey=10, pc_cyl_ptile=25, pc_ptile=99, pc=0):
    otherpcs = list(set(np.arange(kpcs.shape[1])) - {pc})
    pcr = np.sqrt(np.sum(kpcs[:, otherpcs] ** 2, axis=1))
    mask = (pcr < np.percentile(pcr, pc_cyl_ptile)) & (
                np.abs(kpcs[:, pc]) < np.percentile(np.abs(kpcs[:, pc]), pc_ptile))
    mn, mx = np.min(kpcs[mask, pc]), np.max(kpcs[mask, pc])
    bins = np.linspace(mn, mx, nkey)
    idx = np.digitize(kpcs[mask, pc], bins) - 1
    uidx, uidxcnt = np.unique(idx, return_counts=True)
    keyq = []
    for i in uidx[uidxcnt >= 10]:
        kq = meanq(qarr[mask, :][idx == i, :])
        keyq.append(kq)
    return np.array(keyq)


def dualquat(q, t):
    assert q.shape[0] == t.shape[0]
    dq = np.zeros(q.shape, dtype=np.complex128)
    dq.real = q
    dq.imag[:, 1:] = t
    dq.imag = qtimes(dq.imag, dq.real) * 0.5
    return dq


def phi5(r, r2=None):
    if r2 is not None:
        r = r.dot(r2.T)
    return np.linalg.norm(np.eye(3) - r)



import os
import pickle
import numpy as np
import lmdb
import multiprocessing
from PIL import Image
from io import BytesIO
from tqdm import tqdm
import torch
from .mrc_preprocess import mrcs_resize, mrcs_to_int8, window_mask
from . import fft
from . import mrc


def create_lmdb_dataset(image_path_list, lmdb_path,  map_size=1024 ** 4, num_processes=None,
                        chunksize=0, resize=224, is_to_int8=True, window=True, window_r=0.85, generate_ft_data=False,
                        save_raw_data=False):
    """
    并行化创建LMDB数据库。
    处理process_item返回的包含多个条目的列表，并在主进程中按序分配全局key。
    """

    if not os.path.exists(lmdb_path):
        os.makedirs(lmdb_path)
    path_id_data_list = []
    # 初始化LMDB环境
    env = lmdb.open(lmdb_path, map_size=map_size, readonly=False, create=True, max_readers=128)
    protein_id_dict = {}
    protein_id_list = []

    if num_processes is None:
        num_processes = 16
    if chunksize == 0:
        chunksize = 1

    with multiprocessing.Pool(processes=num_processes) as pool:

        protein_id = 0
        # 生成任务列表（索引+路径）
        tasks = [(idx, data_path, resize, is_to_int8, window, window_r, generate_ft_data, save_raw_data) for
                 idx, data_path in enumerate(image_path_list)]

        print(f"Starting parallel processing of {len(tasks)} files...")
        # 使用imap保证结果按任务顺序返回
        results_iterator = pool.imap(lmdb_process_item, tasks, chunksize=chunksize)

        # 使用事务进行写入
        with env.begin(write=True) as txn:
            global_item_index = 0
            # 创建两个进度条：处理文件和写入项目
            with tqdm(total=len(tasks), desc="Preprocessing cryoem data") as pbar_files:
                # tqdm(desc="Writing items to LMDB") as pbar_items:
                for original_idx, path_id_data, item_list in results_iterator:
                    if item_list:  # 如果处理成功且有数据
                        for serialized_data in item_list:
                            key = f"{global_item_index}".encode()
                            txn.put(key, serialized_data)
                            global_item_index += 1
                            # pbar_items.update(1)  # 更新项目进度条
                    path_id_data_list.extend(path_id_data)
                    protein_name = tasks[original_idx][1].split('/')[-3]
                    if protein_name not in protein_id_dict:
                        protein_id_dict[protein_name] = protein_id
                        protein_id += 1
                    protein_id_list.extend([protein_id_dict[protein_name] for _ in range(len(path_id_data))])
                    pbar_files.update(1)  # 更新文件进度条

    env.close()
    save_data_path = os.path.dirname(lmdb_path)
    pickle.dump(protein_id_dict, open(os.path.join(save_data_path, 'protein_id_dict.data'), 'wb'))
    pickle.dump(protein_id_list, open(os.path.join(save_data_path, 'protein_id_list.data'), 'wb'))


    print("LMDB dataset creation finished.")



def lmdb_process_item(args):
    """
    处理单个pickle文件的加载、堆栈分割和序列化。
    支持 pil_image/np_image_raw/np_image_FT 为 [n, d, d] 形状的 images stack。
    不在此函数中生成最终的LMDB key。
    """
    idx, data_path, resize, is_to_int8, window, window_r, generate_ft_data, save_raw_data = args
    path_id_my_data = []
    processed_items_data = []  # 存储序列化后的单个图像数据
    data = {}
    try:
        pil_image = None
        np_image_raw = None
        np_image_FT = None
        n_total = 0  # Initialize to 0

        if data_path.endswith('.data'):
            # Load data from pickle files
            # Load pil_image (processed image)
            if os.path.exists(data_path):
                with open(data_path, 'rb') as f:
                    pil_image = pickle.load(f)
                    if isinstance(pil_image, np.ndarray) and pil_image.ndim >= 2:
                        # Assuming pil_image loaded from .data is the primary processed data
                        processed_mrcs = pil_image
                        if processed_mrcs.ndim == 3:
                            n_total = processed_mrcs.shape[0]
                        elif processed_mrcs.ndim <= 2:  # Handle single image case
                            n_total = 1
                        else:
                            print(
                                f"Warning: Unexpected dimensions for data at {data_path}: {processed_mrcs.ndim}. Skipping.")
                            return (idx, [])  # Return empty list
                    elif isinstance(pil_image, Image.Image):
                        processed_mrcs = np.array(pil_image)
                        n_total = 1
                    else:
                        print(f"Warning: Unexpected data type in {data_path}: {type(pil_image)}. Skipping.")
                        return (idx, [])  # Return empty list


            else:
                print(f"Warning: Processed image file not found at {data_path}. Skipping.")
                return (idx, [])  # Return empty list if the primary file is missing

            # Construct paths for raw and FT data based on assumed directory structure
            pkl_path_parts = data_path.split('/')
            # Ensure path has enough parts to modify the directory
            if len(pkl_path_parts) > 3:
                # Assuming data_path is like .../processed/dataset_name/file_name.data
                # Raw/FT paths would be .../raw/dataset_name/file_name.data
                pkl_path_raw_parts = pkl_path_parts[:-4] + ['raw'] + pkl_path_parts[-3:]
                pkl_path_FT_parts = pkl_path_parts[:-4] + ['FT'] + pkl_path_parts[-3:]

                pkl_path_raw = '/'.join(pkl_path_raw_parts)
                pkl_path_FT = '/'.join(pkl_path_FT_parts)

                # Load raw image data
                if os.path.exists(pkl_path_raw):
                    with open(pkl_path_raw, 'rb') as f:
                        np_image_raw = pickle.load(f)

                # Load FT image data
                if os.path.exists(pkl_path_FT) and generate_ft_data:
                    with open(pkl_path_FT, 'rb') as f:
                        np_image_FT = pickle.load(f)
            else:
                print(f"Warning: Could not construct raw/FT paths for {data_path}. Assuming no raw/FT data.")


        else:  # Assuming MRC file
            try:
                # mrc_data = mrcfile.open(data_path).data
                mrc_data, _ = mrc.parse_mrc(data_path)
                np_image_raw = np.float32(mrc_data)
                imgs_len = np_image_raw.shape[0]  # Assuming [n, d, d]
                n_total = imgs_len  # Number of items is the number of images in stack

                processed_mrcs = np_image_raw.copy()
                if np_image_raw.shape[1] != resize:
                    processed_mrcs = mrcs_resize(processed_mrcs, resize, resize)

                if is_to_int8:
                    processed_mrcs = mrcs_to_int8(processed_mrcs)
                    # processed_mrcs = to_int8(processed_mrcs)

                if window:
                    np_image_raw *= window_mask(np_image_raw.shape[-1], window_r, .99)

                if generate_ft_data:
                    particles = []
                    for i in range(imgs_len):
                        img = np_image_raw[i]
                        particles.append(fft.ht2_center(img))

                    np_image_FT = np.asarray(particles, dtype=np.float32)
                    np_image_FT = fft.symmetrize_ht(np_image_FT)
            except Exception as mrc_e:
                print(f"Error processing MRC file {data_path}: {str(mrc_e)}. Skipping.")
                return (idx, [])

        # --- Process each image in the potential stack ---
        # If n_total is 0 after processing, return empty list
        if n_total == 0:
            print(f"Warning: No valid data found in {data_path}. Skipping.")
            return (idx, [])

        for i in range(n_total):
            current_pil = None
            current_raw = None
            current_FT = None

            path_id_my_data.append(os.path.join(data_path.split('/')[-1], str(i + 1).zfill(6)))

            # Extract processed image slice
            if processed_mrcs is not None:
                if processed_mrcs.ndim == 3 and i < n_total:
                    current_pil = Image.fromarray(processed_mrcs[i]).convert('L')
                elif processed_mrcs.ndim <= 2 and n_total == 1:  # Handle single image case
                    current_pil = Image.fromarray(processed_mrcs).convert('L')
                # else: index out of bounds or wrong dim
                data['image_processed'] = current_pil

            # Extract slice for raw data if it's a stack and index is valid
            if save_raw_data and isinstance(np_image_raw, np.ndarray):
                if np_image_raw.ndim == 3 and i < n_total:
                    current_raw = np_image_raw[i]
                elif np_image_raw.ndim <= 2 and n_total == 1:  # Handles single image case
                    current_raw = np_image_raw
                # else: index out of bounds for this stack type or wrong dim, current_raw remains None
                data['image_raw'] = current_raw

            # Extract slice for FT data if it's a stack and index is valid
            if generate_ft_data and isinstance(np_image_FT, np.ndarray):
                if np_image_FT.ndim == 3 and i < n_total:
                    current_FT = np_image_FT[i]
                elif np_image_FT.ndim <= 2 and n_total == 1:  # Handles single image case
                    current_FT = np_image_FT
                # else: index out of bounds for this stack type or wrong dim, current_FT remains None
                data['image_FT'] = current_FT

            if current_pil is not None or current_raw is not None or current_FT is not None:
                # data = {'image_processed': current_pil, 'image_raw': current_raw, 'image_FT': current_FT}
                buffer = BytesIO()
                torch.save(data, buffer)
                serialized_data = buffer.getvalue()
                # Append only the serialized data value
                processed_items_data.append(serialized_data)
            else:
                print(f"Warning: No data found for slice {i} in file {data_path}. Skipping this slice.")

        # Return the original file index and the list of serialized item data
        return (idx, path_id_my_data, processed_items_data)

    except Exception as e:
        print(f"Error processing {data_path}: {str(e)}")
        # Return original index and empty list to indicate failure for the whole file
        return (idx, [])


